from typing import Optional
import numpy as np
from .activation import Activation


class Tanh(Activation):
    def function(self, x: np.ndarray) -> np.ndarray:
        """
        A method for the tanh function.

        This method implements the tanh function using the np.tanh function:
        (exp(x) - exp(-x)) / (exp(x) + exp(-x))

        Args:
            x: A numpy array of any shape, representing the input to the tanh
            function.

        Returns:
            A numpy array of the same shape as x, representing the output of
            the tanh function.
        """
        return np.tanh(x)

    def derivative(
        self, x: np.ndarray, output: Optional[np.ndarray] = None
    ) -> np.ndarray:
        """
        A method for the derivative of the tanh function.

        This method implements the derivative of the tanh function using the
        np.tanh function: 1 - tanh(x)^2

        Args:
            x: A numpy array of any shape, representing the input to the
            tanh function.

        Returns:
            A numpy array of the same shape as x, representing the output
            of the derivative of the tanh function.
        """
        return 1 - np.tanh(x) ** 2



import numpy as np
import pytest

from inventronet.activations import LeakyReLU


class TestLeakyReLU:
    # Test the LeakyReLU activation function

    @pytest.fixture
    def leaky_relu(self) -> LeakyReLU:
        # Fixture for the LeakyReLU function with default alpha
        yield LeakyReLU()

    @pytest.fixture
    def leaky_relu_01(self):
        # Fixture for the LeakyReLU function with alpha = 0.1
        return LeakyReLU(alpha=0.1)

    @pytest.mark.parametrize(
        "x, y",
        [
            (np.array([0, 1, -1]), np.array([0, 1, -0.01])),
            (np.array([2, -2, 0.5]), np.array([2, -0.02, 0.5])),
        ],
    )
    def test_leaky_relu_function(self, leaky_relu: LeakyReLU, x, y):
        # Test the function method of the LeakyReLU function with different
        # inputs and outputs
        assert np.allclose(leaky_relu.function(x), y)

    @pytest.mark.parametrize(
        "x, y",
        [
            (np.array([0, 1, -1]), np.array([0.01, 1, 0.01])),
            (np.array([2, -2, 0.5]), np.array([1, 0.01, 1])),
        ],
    )
    def test_leaky_relu_derivative(self, leaky_relu: LeakyReLU, x, y):
        # Test the derivative method of the LeakyReLU function with different
        # inputs and outputs
        assert np.allclose(leaky_relu.derivative(x), y)

    @pytest.mark.parametrize(
        "x, y",
        [
            (np.array([0, 1, -1]), np.array([0, 1, -0.1])),
            (np.array([2, -2, 0.5]), np.array([2, -0.2, 0.5])),
        ],
    )
    def test_leaky_relu_function_01(self, leaky_relu_01: LeakyReLU, x, y):
        # Test the function method of the LeakyReLU function with different
        # inputs and outputs
        assert np.allclose(leaky_relu_01.function(x), y)

    @pytest.mark.parametrize(
        "x, y",
        [
            (np.array([0, 1, -1]), np.array([0.1, 1, 0.1])),
            (np.array([2, -2, 0.5]), np.array([1, 0.1, 1])),
        ],
    )
    def test_leaky_relu_derivative_01(self, leaky_relu_01: LeakyReLU, x, y):
        # Test the derivative method of the LeakyReLU function with different
        # inputs and outputs
        assert np.allclose(leaky_relu_01.derivative(x), y)



import numpy as np
import pytest

from inventronet.preprocessing import train_test_split


@pytest.fixture(scope="module")
def data():
    X = np.random.rand(100, 5)
    y = np.random.randint(0, 2, size=(100,))
    return X, y


@pytest.fixture(scope="module")
def splitted_data(data):
    X, y = data
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)
    return X_train, X_test, y_train, y_test


def test_X_train_shape(splitted_data):
    X_train, _, _, _ = splitted_data
    assert X_train.shape == (80, 5)


def test_X_test_shape(splitted_data):
    _, X_test, _, _ = splitted_data
    assert X_test.shape == (20, 5)


def test_y_train_shape(splitted_data):
    _, _, y_train, _ = splitted_data
    assert y_train.shape == (80,)


def test_y_test_shape(splitted_data):
    _, _, _, y_test = splitted_data
    assert y_test.shape == (20,)


def test_indices_shuffled(data, splitted_data):
    X, _ = data
    X_train, X_test, _, _ = splitted_data
    assert not np.array_equal(np.arange(100), np.concatenate([X_train, X_test]))



import numpy as np
import pytest

from inventronet.activations import Tanh


class TestTanh:
    # Test the Tanh activation function

    @pytest.fixture
    def tanh(self) -> Tanh:
        # Fixture for the Tanh function
        yield Tanh()

    @pytest.mark.parametrize(
        "x, y",
        [
            (np.array([0, 1, -1]), np.array([0, 0.7615942, -0.7615942])),
            (np.array([2, -2, 0.5]), np.array([0.9640276, -0.9640276, 0.4621172])),
        ],
    )
    def test_tanh_function(self, tanh: Tanh, x, y):
        # Test the function method of the Tanh function with different
        # inputs and outputs
        assert np.allclose(tanh.function(x), y, atol=1e-4)

    @pytest.mark.parametrize(
        "x, y",
        [
            (np.array([0, 1, -1]), np.array([1, 0.4199743, 0.4199743])),
            (np.array([2, -2, 0.5]), np.array([0.0706508, 0.0706508, 0.7864477])),
        ],
    )
    def test_tanh_derivative(self, tanh: Tanh, x, y):
        # Test the derivative method of the Tanh function with different
        # inputs and outputs
        assert np.allclose(tanh.derivative(x), y, atol=1e-4)



import numpy as np
from .loss import Loss


class MeanSquaredError(Loss):
    def function(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        """
        Computes the mean squared error loss between the true labels
        and the predicted values.

        Parameters
        ----------
        y_true : np.ndarray
            The true labels of the data, encoded as one-hot vectors or
              as scalars.
            Shape: (n_samples, n_classes) or (n_samples,)
        y_pred : np.ndarray
            The predicted values of the data, output of the linear function
            or the softmax function.
            Shape: (n_samples, n_classes) or (n_samples,)

        Returns
        -------
        float
            The mean squared error loss, averaged over all samples and
            classes or samples.
        """
        return np.mean((y_true - y_pred) ** 2)

    def gradient(self, y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
        """
        Computes the derivative of the mean squared error loss with respect
        to the predictions.

        Parameters
        ----------
        y_true : np.ndarray
            The true labels of the data, encoded as one-hot vectors or
            as scalars.
            Shape: (n_samples, n_classes) or (n_samples,)
        y_pred : np.ndarray
            The predicted values of the data, output of the linear
            function or the softmax function.
            Shape: (n_samples, n_classes) or (n_samples,)

        Returns
        -------
        np.ndarray
            The derivative of the mean squared error loss, scaled
            by the number of samples.
            Shape: (n_samples, n_classes) or (n_samples,)
        """
        return 2 * (y_pred - y_true) / y_true.size



import numpy as np
import pytest

from inventronet.activations import Linear


class TestLinear:
    # Test the Linear activation function

    @pytest.fixture
    def linear(self) -> Linear:
        # Fixture for the Linear function
        yield Linear()

    @pytest.mark.parametrize(
        "x, y",
        [
            (np.array([0, 1, -1]), np.array([0, 1, -1])),
            (np.array([2, -2, 0.5]), np.array([2, -2, 0.5])),
        ],
    )
    def test_linear_function(self, linear: Linear, x, y):
        # Test the function method of the Linear function with different
        # inputs and outputs
        assert np.allclose(linear.function(x), y)

    @pytest.mark.parametrize(
        "x, y",
        [
            (np.array([0, 1, -1]), np.array([1, 1, 1])),
            (np.array([2, -2, 0.5]), np.array([1, 1, 1])),
        ],
    )
    def test_linear_derivative(self, linear: Linear, x, y):
        # Test the derivative method of the Linear function with different
        # inputs and outputs
        assert np.allclose(linear.derivative(x), y)



import inspect
from typing import Type

import numpy as np
import pytest
from _pytest.fixtures import SubRequest

from inventronet import optimizers as optimizers_module
from inventronet.activations import Sigmoid
from inventronet.activations.activation import Activation
from inventronet.optimizers.optimizer import Optimizer


def generate_random_array(shape):
    return np.random.randn(*shape)


@pytest.fixture
def random_input() -> np.ndarray:
    yield generate_random_array((10, 5))


@pytest.fixture
def input_dim(random_input: np.ndarray) -> int:
    yield random_input.shape[1]


@pytest.fixture
def random_error(random_input: np.ndarray) -> np.ndarray:
    yield generate_random_array(random_input.shape) / 2


@pytest.fixture
def random_biases(random_error: np.ndarray) -> np.ndarray:
    yield generate_random_array((random_error.shape[1],)) * 2


all_optimizers = [
    optimizer
    for _, optimizer in inspect.getmembers(optimizers_module, inspect.isclass)
    if issubclass(optimizer, Optimizer) and optimizer != Optimizer
]


@pytest.fixture(params=all_optimizers)
def optimizer(request: SubRequest) -> Type[Optimizer]:
    opt: Type[Optimizer] = request.param()
    # Increase the learning rate
    opt.learning_rate = 0.1
    yield opt


@pytest.fixture
def activation() -> Type[Activation]:
    yield Sigmoid()



from abc import ABC, abstractmethod


class Metric(ABC):
    @abstractmethod
    def call(self, y_true, y_pred) -> float:
        raise NotImplementedError("The call method must be implemented.")

    def __call__(self, y_true, y_pred) -> float:
        return self.call(y_true, y_pred)



from typing import Dict

import numpy as np

from inventronet.optimizers import StochasticGradientDescent


class TestGradientDescent:
    def test_gradient_descent_init_default(self):
        optimizer = StochasticGradientDescent()
        assert optimizer.learning_rate == 0.01

    def test_gradient_descent_init_custom(self):
        optimizer = StochasticGradientDescent(learning_rate=0.05)
        assert optimizer.learning_rate == 0.05

    def test_gradient_descent_update(
            self, params: Dict[str, np.ndarray], gradients: Dict[str, np.ndarray]
    ):
        optimizer = StochasticGradientDescent(learning_rate=0.1)

        optimizer.update(0, params, gradients)

        # Expected updated parameters after applying gradient descent
        expected_params = {
            "weights": np.array([[0.99, 1.98], [2.97, 3.96]]),
            "biases": np.array([0.49, 0.51]),
        }

        # Check if the updated parameters are equal to the expected values
        assert np.allclose(params["weights"], expected_params["weights"])
        assert np.allclose(params["biases"], expected_params["biases"])

    def test_gradient_descent_update_with_zeros(
            self, zero_params: Dict[str, np.ndarray], zero_gradients: Dict[str, np.ndarray]
    ):
        optimizer = StochasticGradientDescent(learning_rate=0.1)

        optimizer.update(0, zero_params, zero_gradients)

        # Expected updated parameters after applying gradient descent
        # with zero gradients
        expected_params = {
            "weights": np.zeros((2, 2)),
            "biases": np.zeros(2),
        }

        # Check if the updated parameters are equal to the expected values
        assert np.allclose(zero_params["weights"], expected_params["weights"])
        assert np.allclose(zero_params["biases"], expected_params["biases"])



import numpy as np
import pytest

from inventronet.losses import MeanSquaredError


class TestMSE:
    # Create an instance of the MSE loss function
    @pytest.fixture
    def mse(self) -> MeanSquaredError:
        yield MeanSquaredError()

    # Test the function method of the MSE loss function with different
    # inputs and outputs
    @pytest.mark.parametrize(
        "y_true, y_pred, expected",
        [
            (np.array([1, 2, 3]), np.array([1, 2, 3]), 0.0),  # exact match
            # one unit difference
            (np.array([1, 2, 3]), np.array([2, 3, 4]), 1.0),
            (
                    np.array([1, 2, 3]),
                    np.array([0, 0, 0]),
                    4.666666666666667,
            ),  # large difference
        ],
    )
    def test_mse_function(self, y_true, y_pred, expected, mse: MeanSquaredError):
        # Compare the expected and actual outputs with a precision
        # of 7 decimal places
        np.testing.assert_almost_equal(
            mse.function(y_true, y_pred), expected, decimal=7
        )

    # Test the derivative method of the MSE loss function with different
    # inputs and outputs
    @pytest.mark.parametrize(
        "y_true, y_pred, expected",
        [
            (
                    np.array([1, 2, 3]),
                    np.array([1, 2, 3]),
                    np.array([0, 0, 0]),
            ),  # exact match
            (
                    np.array([1, 2, 3]),
                    np.array([2, 3, 4]),
                    np.array([0.66666667, 0.66666667, 0.66666667]),
            ),  # one unit difference
            (
                    np.array([1, 2, 3]),
                    np.array([0, 0, 0]),
                    np.array([-0.66666667, -1.33333333, -2.0]),
            ),  # large difference
        ],
    )
    def test_mse_derivative(self, y_true, y_pred, expected, mse: MeanSquaredError):
        # Compare the expected and actual outputs with a precision of
        # 7 decimal places
        np.testing.assert_almost_equal(
            mse.gradient(y_true, y_pred), expected, decimal=7
        )



from typing import Dict

import numpy as np

from inventronet.optimizers import Adam


class TestAdam:
    def test_adam_init_default(self):
        optimizer = Adam()
        assert optimizer.learning_rate == 0.001
        assert optimizer.beta1 == 0.9
        assert optimizer.beta2 == 0.999
        assert optimizer.epsilon == 1e-8

    def test_adam_init_custom(self):
        optimizer = Adam(learning_rate=0.01, beta1=0.8, beta2=0.99, epsilon=1e-7)
        assert optimizer.learning_rate == 0.01
        assert optimizer.beta1 == 0.8
        assert optimizer.beta2 == 0.99
        assert optimizer.epsilon == 1e-7

    def test_adam_update(
            self, params: Dict[str, np.ndarray], gradients: Dict[str, np.ndarray]
    ):
        optimizer = Adam(learning_rate=0.001)

        optimizer.update(0, params, gradients)

        # Corrected expected updated parameters after applying Adam optimizer (calculated manually)
        expected_params = {
            "weights": np.array([[0.999, 1.999], [2.999, 3.999]]),
            "biases": np.array([0.499, 0.501]),
        }

        # Check if the updated parameters are equal to the expected values
        assert np.allclose(
            params["weights"], expected_params["weights"], rtol=1e-7, atol=1e-7
        )
        assert np.allclose(
            params["biases"], expected_params["biases"], rtol=1e-7, atol=1e-7
        )

    def test_adam_update_with_zeros(
            self, zero_params: Dict[str, np.ndarray], zero_gradients: Dict[str, np.ndarray]
    ):
        optimizer = Adam(learning_rate=0.01)

        optimizer.update(0, zero_params, zero_gradients)

        # Expected updated parameters after applying Adam optimizer with
        # zero gradients
        expected_params = {
            "weights": np.zeros((2, 2)),
            "biases": np.zeros(2),
        }

        # Check if the updated parameters are equal to the expected values
        assert np.allclose(zero_params["weights"], expected_params["weights"], rtol=1e-7)
        assert np.allclose(zero_params["biases"], expected_params["biases"], rtol=1e-7)



from typing import Tuple
import numpy as np


from .layer import Layer


class BatchNormalization(Layer):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        momentum: float = 0.9,
        epsilon: float = 1e-5,
    ) -> None:
        """Initialize the BatchNormalization layer with the given arguments.

        Args:
            input_dim: The dimension of the input features.
            output_dim: The dimension of the output features.
            momentum: The momentum for the moving average of the running
            statistics. Defaults to 0.9.
            epsilon: The small constant for numerical stability.
              Defaults to 1e-5.
        """
        super().__init__(input_dim, output_dim, None)
        self.normalized_input = None
        self.batch_var = None
        self.batch_mean = None
        self.gamma = np.ones((output_dim,))
        self.beta = np.zeros((output_dim,))
        self.running_mean = np.zeros((output_dim,))
        self.running_var = np.ones((output_dim,))
        self.momentum = momentum
        self.epsilon = epsilon
        self.parameters = {"gamma": self.gamma, "beta": self.beta}
        self.gradients = {
            "gamma": np.zeros_like(self.gamma),
            "beta": np.zeros_like(self.beta),
        }

    def forward(self, inputs: np.ndarray, training: bool = False) -> np.ndarray:
        """Perform the layer operation on the inputs.

        Args:
            inputs: The input array of shape (batch_size, input_dim).
            training: The training flag, True or False.

        Returns:
            The output array of shape (batch_size, output_dim).
        """
        self.previous_layer_output = inputs
        if training:
            batch_mean = np.mean(inputs, axis=0)
            batch_var = np.var(inputs, axis=0)
            self.running_mean = (
                self.momentum * self.running_mean + (1 - self.momentum) * batch_mean
            )
            self.running_var = (
                self.momentum * self.running_var + (1 - self.momentum) * batch_var
            )
        else:
            batch_mean = self.running_mean
            batch_var = self.running_var
        normalized_input = (inputs - batch_mean) / np.sqrt(batch_var + self.epsilon)
        self.batch_mean = batch_mean
        self.batch_var = batch_var
        self.normalized_input = normalized_input
        output = self.gamma * normalized_input + self.beta
        return output

    def backward(
        self,
        error: np.ndarray,
        prev_output: np.ndarray = None,
        training: bool = False,
    ) -> np.ndarray:
        """Perform the backward propagation on the layer.

        Args:
            training: A boolean indicating whether the layer is in training
            error: The error array of shape (batch_size, output_dim).
            prev_output: The previous layer output, not used in this layer.

        Returns:
            The gradient array of shape (batch_size, input_dim).
        """
        batch_size = error.shape[0]
        grad_gamma = np.sum(error * self.normalized_input, axis=0)
        grad_beta = np.sum(error, axis=0)

        self.gradients["gamma"] = grad_gamma
        self.gradients["beta"] = grad_beta

        grad_normalized_input = error * self.gamma
        grad_input = (
            batch_size * grad_normalized_input
            - np.sum(grad_normalized_input, axis=0)
            - self.normalized_input
            * np.sum(grad_normalized_input * self.normalized_input, axis=0)
        ) / (batch_size * np.sqrt(self.batch_var + self.epsilon))
        grad_prev_output = grad_input
        return grad_prev_output

    def get_parameters(self) -> Tuple[np.ndarray, np.ndarray]:
        """Return the weights and biases of the layer as numpy arrays."""
        return self.gamma, self.beta



from .optimizer import Optimizer


class StochasticGradientDescent(Optimizer):
    def __init__(self, learning_rate: float = 0.01):
        self.learning_rate = learning_rate

    def update(self, layer_id: int, params: dict, gradients: dict):
        for key in params.keys():
            params[key] -= self.learning_rate * gradients[key]



from abc import ABC, abstractmethod
from typing import List, Tuple, Type

import numpy as np

from ..layers.layer import Layer
from ..losses.loss import Loss
from ..metrics.metric import Metric
from ..optimizers.optimizer import Optimizer


# Define an abstract base class for a model
class Model(ABC):
    # Define the initialization method
    def __init__(self) -> None:
        """Initialize an empty list of layers."""
        # Initialize an empty list of layers
        self.layers: List[Type[Layer]] = []

    # Define an abstract method for adding a layer to the model
    @abstractmethod
    def add(self, layer: Layer) -> None:
        """Add a layer to the model.

        Args:
            layer (Layer): The layer to be added.
        """
        raise NotImplementedError("The add method must be implemented.")

    # Define an abstract method for compiling the model with a loss
    # function and a metric
    @abstractmethod
    def compile(
            self, loss: Loss, optimizer: Optimizer, metrics: List[Type[Metric]]
    ) -> None:
        """Compile the model with a loss function and a metric.

        Args:
            loss (Loss): The loss function to be used.
            optimizer (Optimizer): The optimizer to be used.
            metrics (Metric): The metrics to be used.
        """
        raise NotImplementedError("The compile method must be implemented.")

    # Define an abstract method for fitting the model on training data
    @abstractmethod
    def fit(
            self,
            x_train: np.ndarray,
            y_train: np.ndarray,
            epochs: int,
    ) -> None:
        """Fit the model on training data.

        Args:
            x_train (np.ndarray): The input data for training.
            y_train (np.ndarray): The output data for training.
            epochs (int): The number of epochs to train the model.
        """
        raise NotImplementedError("The fit method must be implemented.")

    # Define an abstract method for predicting the output for new data
    @abstractmethod
    def predict(self, x_test: np.ndarray) -> np.ndarray:
        """Predict the output for new data.

        Args:
            x_test (np.ndarray): The input data for testing.

        Returns:
            np.ndarray: The output data for testing.
        """
        raise NotImplementedError("The predict method must be implemented.")

    # Define an abstract method for evaluating the model on test data
    @abstractmethod
    def evaluate(self, x_test: np.ndarray, y_test: np.ndarray) -> Tuple[float, float]:
        """Evaluate the model on test data.

        Args:
            x_test (np.ndarray): The input data for testing.
            y_test (np.ndarray): The output data for testing.

        Returns:
            Tuple[float, float]: The loss and the metric values for the test data.
        """
        raise NotImplementedError("The evaluate method must be implemented.")



# Define the precision metric
import numpy as np
from .metric import Metric


class Precision(Metric):
    def call(self, y_true, y_pred) -> float:
        # Round the predictions to get the binary labels
        y_pred = np.round(y_pred)
        # Count the true positives and the predicted positives
        tp = np.sum(y_true * y_pred)
        pp = np.sum(y_pred)
        # Compute the precision
        return tp / pp if pp > 0.0 else 0.0



from typing import Optional
import numpy as np
from .activation import Activation


class SoftMax(Activation):
    """
    A class for the softmax activation function.

    The softmax activation function is defined as:
    softmax(x)_i = exp(x_i) / sum(exp(x_j)) for j = 1, ..., n

    It is a vector-wise function that normalizes the input vector to a
    probability distribution, where the sum of the output elements is 1.
    It is often used as the output layer of a neural network for
    multi-class classification problems.

    Attributes:
        None
    """

    # Define the function method
    def function(self, x: np.ndarray) -> np.ndarray:
        """
        A method for the function.

        This method implements the softmax function using the np.exp and
        np.sum functions:
        softmax(x)_i = exp(x_i) / sum(exp(x_j)) for j = 1, ..., n

        Args:
            x: A numpy array of shape (n,) or (m, n), representing the
            input to the softmax function.

        Returns:
            A numpy array of the same shape as x, representing the
            output of the softmax function.
        """
        exp_x = np.exp(x - np.max(x, axis=-1, keepdims=True))
        sum_exp_x = np.sum(exp_x, axis=-1, keepdims=True)
        return exp_x / sum_exp_x

    # Define the derivative method
    def derivative(
        self, x: np.ndarray, output: Optional[np.ndarray] = None
    ) -> np.ndarray:
        """
        A method for the derivative.

        This method implements the derivative of the softmax function using
        the np.outer and np.exp functions:
        softmax(x)_i * (delta_i_j - softmax(x)_j)

        Args:
            x: A numpy array of shape (n,) or (m, n), representing the input
            to the softmax function.

        Returns:
            A numpy array of shape (n, n) or (m, n, n), representing the
            output of the derivative of the softmax function.
        """
        if output is None:
            output = self.function(x)

        outer_product = np.outer(output, output)
        outer_product = outer_product.reshape(*x.shape, *x.shape[-1:])
        diag_output = np.identity(x.shape[-1]) * output[..., np.newaxis]
        return diag_output - outer_product



import numpy as np
import pytest

from inventronet.metrics import Precision


@pytest.fixture
def precision() -> Precision:
    yield Precision()


@pytest.mark.parametrize(
    "y_true, y_pred, expected",
    [
        (
            np.array([1, 0, 0, 1, 1]),
            np.array([0.9, 0.1, 0.2, 0.8, 0.7]),
            1.0,
        ),  # exact match
        (
            np.array([1, 0, 0, 1, 1]),
            np.array([0.6, 0.4, 0.3, 0.7, 0.5]),
            1.0,
        ),  # small difference
        (
            np.array([1, 0, 0, 1, 1]),
            np.array([0.1, 0.9, 0.8, 0.2, 0.3]),
            0.0,
        ),  # large difference
        (
            np.array([1, 0, 0, 1, 1]),
            np.array([0.6, 0.6, 0.6, 0.6, 0.6]),
            0.6,
        ),  # equal predictions
        (
            np.array([1, 0, 0, 1, 1]),
            np.array([0.0, 0.0, 0.0, 0.0, 0.0]),
            0.0,
        ),  # no predictions
        (
            np.array([1, 1, 1, 1, 1]),
            np.array([0.9, 0.9, 0.9, 0.9, 0.9]),
            1.0,
        ),  # all true positives and no false positives
        (
            np.array([0, 0, 0, 0, 0]),
            np.array([0.9, 0.9, 0.9, 0.9, 0.9]),
            0.0,
        ),  # no true positives and all false positives
        (
            [1, 0, 1, 0, 1],
            [0.9, 0.1, 0.9, 0.1, 0.9],
            1.0,
        ),  # Python list input
        (
            np.array([]),
            np.array([]),
            0.0,
        ),  # empty arrays
    ],
)
def test_precision_call(y_true, y_pred, expected, precision: Precision):
    # Compare the expected and actual outputs with a precision of 7 decimal
    # places
    np.testing.assert_almost_equal(precision.call(y_true, y_pred), expected, decimal=7)



from typing import Type, Any

import numpy as np
import pytest
from _pytest.capture import CaptureResult

from inventronet.activations import Sigmoid
from inventronet.activations.activation import Activation
from inventronet.layers import Dense
from inventronet.optimizers.optimizer import Optimizer


class TestDense:
    @pytest.fixture
    def dense(
        self, activation: Type[Activation], input_dim: int, random_error: np.ndarray
    ) -> Dense:
        yield Dense(
            input_dim,
            random_error.shape[1],
            activation=activation,
            use_bias=False,
        )

    @pytest.fixture
    def dense_with_bias(self, dense: Dense, random_biases: np.ndarray) -> Dense:
        dense.use_bias = True
        dense.biases = random_biases
        yield dense

    @pytest.fixture
    def forward_output_train(
        self, random_input: np.ndarray[Any, np.float64], dense: Dense
    ) -> np.ndarray:
        yield dense.forward(random_input)

    def test_linearity(self, dense: Dense, random_input: np.ndarray):
        """
        Test the linearity property of the Dense layer without
        activation function. Ensures the Dense layer performs
        a linear transformation when no activation function is used.
        """
        dense.activation = None
        input1, input2 = random_input[:5], random_input[5:]
        alpha, beta = 0.3, 0.7
        linear_combination = alpha * input1 + beta * input2
        output1, output2 = dense.forward(input1), dense.forward(input2)
        expected_output = alpha * output1 + beta * output2
        linear_combination_output = dense.forward(linear_combination)
        assert np.allclose(linear_combination_output, expected_output, atol=1e-6)

    def test_fully_connected(self, dense: Dense, random_input: np.ndarray):
        """
        Test the fully connected property of the Dense layer. Ensures
        that every input node is connected to every output node.
        """
        output = dense.forward(random_input)
        for input_vector, output_vector in zip(random_input, output):
            assert not np.all(output_vector == 0) and not np.all(input_vector == 0)

    def test_differentiability(
        self, dense: Dense, random_input: np.ndarray, random_error: np.ndarray
    ):
        """
        Test the differentiability property of the Dense layer. Ensures
        that forward and backward passes can be performed without errors.
        """
        dense.forward(random_input)
        try:
            dense.backward(random_error)
        except Exception as e:
            pytest.fail(f"Differentiability test failed with exception: {e}")

    def test_weight_gradients(
        self, dense: Dense, random_input: np.ndarray, random_error: np.ndarray
    ):
        """
        Test the weight gradient computation after a backward pass. Ensures
        that the Dense layer's weight gradients are computed after backward propagation.
        """
        dense.forward(random_input)
        dense.backward(random_error)
        assert "weights" in dense.gradients
        assert dense.gradients["weights"].shape == dense.weights.shape

    def test_bias_gradients(
        self, dense_with_bias: Dense, random_input: np.ndarray, random_error: np.ndarray
    ):
        """
        Test the bias gradient computation after a backward pass. Ensures
        that the Dense layer's bias gradients are computed after backward propagation.
        """
        dense_with_bias.forward(random_input)
        dense_with_bias.backward(random_error)
        assert "biases" in dense_with_bias.gradients
        assert dense_with_bias.gradients["biases"].shape == dense_with_bias.biases.shape

    class TestDenseInit:
        def test_name(self, dense: Dense):
            assert dense.name == "Dense"

        def test_input_shape(self, dense: Dense):
            assert dense.input_shape == (5,)

        def test_output_shape(self, dense: Dense, random_error: np.ndarray):
            assert dense.output_shape == (random_error.shape[1],)

        def test_summary(
            self, capsys: pytest.CaptureFixture, dense: Dense, random_error: np.ndarray
        ):
            expected_output = f"Layer: Dense\nInput shape: (5,)\nOutput shape: ({random_error.shape[1]},)\nNumber of parameters: 25\n"
            dense.summary()
            captured: CaptureResult = capsys.readouterr()
            assert captured.out == expected_output

    @pytest.fixture
    def forward_output_with_bias(
        self,
        random_input: np.ndarray[Any, np.float64],
        dense_with_bias: Dense,
    ) -> np.ndarray:
        yield dense_with_bias.forward(random_input)

    def test_forward(
        self,
        random_input: np.ndarray[Any, np.float64],
        activation: Sigmoid,
        forward_output_train: np.ndarray,
        dense: Dense,
    ):
        expected_output = activation(np.dot(random_input, dense.weights))
        assert np.allclose(forward_output_train, expected_output)

    def test_forward_invalid_input(
        self, random_input: np.ndarray[Any, np.float64], dense: Dense
    ):
        invalid_input = random_input[:, :4]
        with pytest.raises(ValueError):
            dense.forward(invalid_input)

    class TestBackward:
        def test_backward_output(
            self,
            random_error: np.ndarray,
            activation: Sigmoid,
            dense: Dense,
            forward_output_train: np.ndarray,
        ):
            expected_output: np.ndarray = np.dot(
                random_error * activation.derivative(forward_output_train),
                dense.weights.T,
            )
            # Assert that the output of the backward method is close to the
            # expected output
            assert np.allclose(dense.backward(random_error), expected_output, atol=1e-3)

    class TestBackwardUpdateWeightAndBiases:
        def test_updated_weights(
            self,
            random_input: np.ndarray,
            random_error: np.ndarray,
            activation: Sigmoid,
            dense_with_bias: Dense,
            forward_output_with_bias: np.ndarray,
            optimizer: Type[Optimizer],
        ):
            original_weights = dense_with_bias.weights.copy()
            dense_with_bias.backward(random_error)
            optimizer.update(0, dense_with_bias.parameters, dense_with_bias.gradients)
            learning_rate = optimizer.learning_rate
            expected_weights: np.ndarray = original_weights - learning_rate * np.dot(
                random_input.T,
                random_error * activation.derivative(forward_output_with_bias),
            )

            assert np.allclose(dense_with_bias.weights, expected_weights, atol=1e-1)

        def test_updated_biases(
            self,
            random_error: np.ndarray,
            activation: Sigmoid,
            dense_with_bias: Dense,
            forward_output_with_bias: np.ndarray,
            optimizer: Type[Optimizer],
        ):
            original_bias = dense_with_bias.biases.copy()
            dense_with_bias.backward(random_error)
            optimizer.update(0, dense_with_bias.parameters, dense_with_bias.gradients)
            expected_bias: np.ndarray = original_bias - 0.01 * np.sum(
                random_error * activation.derivative(forward_output_with_bias), axis=0
            )
            assert np.allclose(dense_with_bias.biases, expected_bias, atol=1e-2)

    @pytest.mark.usefixtures("forward_output_train")
    def test_backward_invalid_error(
        self, random_error: np.ndarray[Any, np.dtype[np.floating[Any]]], dense: Dense
    ):
        invalid_error = random_error[:, :2]
        with pytest.raises(ValueError):
            dense.backward(invalid_error, 0.01)



from abc import ABC, abstractmethod
from typing import Tuple, Type

import numpy as np

from ..activations.activation import Activation


class Layer(ABC):

    @abstractmethod
    def __init__(
            self,
            input_dim: int,
            output_dim: int,
            activation: Type[Activation],
    ) -> None:
        """Initialize the layer with the given arguments.

        Args:
            input_dim: The dimension of the input features.
            output_dim: The dimension of the output features.
            activation: The activation function for the layer.
        """
        self.input_dim: int = input_dim
        self.output_dim: int = output_dim
        self.activation: Type[Activation] = activation
        self.parameters = None
        self.gradients = None

        self.previous_layer_output = None

    @abstractmethod
    def forward(self, inputs: np.ndarray, training: bool = False) -> np.ndarray:
        """Perform the layer operation on the inputs.

        Args:
            training: A boolean indicating whether the layer is in training
            inputs: The input array of shape (batch_size, input_dim).

        Returns:
            The output array of shape (batch_size, output_dim).
        """
        raise NotImplementedError("Layer.forward is not implemented.")

    @abstractmethod
    def backward(
            self,
            grad: np.ndarray,
            prev_output: np.ndarray = None,
            training: bool = False,
    ) -> np.ndarray:
        """Perform the backward propagation on the layer.

        Args:
            training: A boolean indicating whether the layer is in training
            grad: The gradient array of shape (batch_size, output_dim).
            prev_output: The previous output array of
                shape (batch_size, input_dim).

        Returns:
            The propagated error array of shape (batch_size, input_dim).
        """
        raise NotImplementedError("Layer.backward is not implemented.")

    @abstractmethod
    def get_parameters(self) -> Tuple[np.ndarray, np.ndarray]:
        """Return the weights and biases of the layer as numpy arrays."""
        raise NotImplementedError("Layer.get_parameters is not implemented.")

    @property
    def name(self) -> str:
        """Return the name of the layer.

        Returns:
            The name of the layer as a string.
        """
        return self.__class__.__name__

    @property
    def input_shape(self) -> tuple:
        """Return the input shape of the layer.

        Returns:
            The input shape of the layer as a tuple of integers.
        """
        return (self.input_dim,)

    @property
    def output_shape(self) -> tuple:
        """Return the output shape of the layer.

        Returns:
            The output shape of the layer as a tuple of integers.
        """
        return (self.output_dim,)

    def summary(self) -> None:
        """Print the summary of the layer.

        Prints the layer name, input shape, output shape, and number of
             parameters.
        """
        print(f"Layer: {self.name}")
        print(f"Input shape: {self.input_shape}")
        print(f"Output shape: {self.output_shape}")
        if self.parameters is not None:
            print(
                "Number of parameters: "
                + f"{sum(param.size for param in self.parameters.values())}"
            )



from typing import Callable, Dict, Tuple, Type
import numpy as np

from ..activations.activation import Activation
from .layer import Layer


class Dense(Layer):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        activation: Type[Activation] = None,
        use_bias: bool = True,
        weight_initializer: Callable = np.random.uniform,
        bias_initializer: Callable = np.zeros,
    ) -> None:
        """Initialize the dense layer with the given arguments.

        Args:
            input_dim: The dimension of the input features.
            output_dim: The dimension of the output features.
            activation: The activation function for the layer.
                Defaults to None.
            use_bias: Whether to use a bias term for the layer.
                Defaults to True.
            weight_initializer: The function to initialize the weight matrix.
                Defaults to np.random.uniform.
            bias_initializer: The function to initialize the bias vector.
                Defaults to np.zeros.
        """
        super().__init__(input_dim, output_dim, activation)
        self.pre_activation_output = None
        self.use_bias: bool = use_bias
        self.weight_initializer: Callable = weight_initializer
        self.bias_initializer: Callable = bias_initializer

        self.weights: np.ndarray = 0.01 * self.weight_initializer(
            size=(self.input_dim, self.output_dim)
        )

        if self.use_bias:
            self.biases: np.ndarray = self.bias_initializer(shape=(self.output_dim,))
        self.parameters = {"weights": self.weights}
        if self.use_bias:
            self.parameters["biases"] = self.biases

        self.gradients: Dict[str, np.ndarray] = {"weights": None, "biases": None}

    def forward(self, inputs: np.ndarray, training: bool = False) -> np.ndarray:
        """Perform the layer operation on the inputs.

        Args:
            training: A boolean indicating whether the layer is in training
            inputs: The input array of shape (batch_size, input_dim).

        Returns:
            The output array of shape (batch_size, output_dim).
        """
        self.previous_layer_output: np.ndarray = inputs

        output = inputs @ self.weights

        if self.use_bias:
            output += self.biases

        if self.activation is not None:
            self.pre_activation_output = output
            return self.activation(output)
        else:
            return output

    def backward(
            self,
            error: np.ndarray,
            prev_output: np.ndarray = None,
            training: bool = False,
    ) -> np.ndarray:
        if prev_output is None:
            prev_output = self.previous_layer_output
        else:
            self.previous_layer_output = prev_output

        if self.activation:
            activation_derivative = self.activation.derivative(
                self.pre_activation_output
            )
            if error.shape == activation_derivative.shape:
                error = error * activation_derivative
            elif len(error.shape) == 2 and len(activation_derivative.shape) == 3:
                error = np.einsum("ij,jk->ik", error, activation_derivative)
            elif len(error.shape) == 1 and len(activation_derivative.shape) == 2:
                error = np.dot(error, activation_derivative)
            else:
                raise ValueError(
                    "The shapes of the error and the activation derivative "
                    + "are incompatible."
                )

        weight_gradient = prev_output.T @ error

        if self.use_bias:
            bias_gradient = np.sum(error, axis=0)
        else:
            bias_gradient = None

        input_gradient = error @ self.weights.T

        self.gradients["weights"] = weight_gradient
        if self.use_bias:
            self.gradients["biases"] = bias_gradient

        return input_gradient

    def get_parameters(self) -> Tuple[np.ndarray, np.ndarray]:
        """Return the weights and biases of the dense layer as numpy arrays."""
        return self.weights.copy(), self.biases.copy()



import numpy as np
import pytest

from inventronet.activations import ReLU


class TestReLU:
    # Test the ReLU activation function

    @pytest.fixture
    def relu(self) -> ReLU:
        # Fixture for the ReLU function
        yield ReLU()

    @pytest.mark.parametrize(
        "x, y",
        [
            (np.array([0, 1, -1]), np.array([0, 1, 0])),
            (np.array([2, -2, 0.5]), np.array([2, 0, 0.5])),
        ],
    )
    def test_relu_function(self, relu: ReLU, x, y):
        # Test the function method of the ReLU function with different
        # inputs and outputs
        assert np.allclose(relu.function(x), y)

    @pytest.mark.parametrize(
        "x, y",
        [
            (np.array([0, 1, -1]), np.array([0, 1, 0])),
            (np.array([2, -2, 0.5]), np.array([1, 0, 1])),
        ],
    )
    def test_relu_derivative(self, relu: ReLU, x, y):
        # Test the derivative method of the ReLU function with different
        # inputs and outputs
        assert np.allclose(relu.derivative(x), y)



from typing import Optional
import numpy as np
from .activation import Activation


class LeakyReLU(Activation):
    def __init__(self, alpha: float = 0.01) -> None:
        """
        A method to initialize the LeakyReLU object with a given alpha
        parameter.

        Args:
            alpha: A float value, representing the slope of the negative part
            of the LeakyReLU function. Default is 0.01.

        Returns:
            None
        """
        super().__init__()
        self.alpha = alpha

    def function(self, x: np.ndarray) -> np.ndarray:
        """
        A method for the LeakyReLU function.

        This method implements the LeakyReLU function using the np.where
        function: max(alpha * x, x)

        Args:
            x: A numpy array of any shape, representing the input to the
            LeakyReLU function.

        Returns:
            A numpy array of the same shape as x, representing the output
            of the LeakyReLU function.
        """
        return np.where(x > 0, x, self.alpha * x)

    def derivative(
        self, x: np.ndarray, output: Optional[np.ndarray] = None
    ) -> np.ndarray:
        """
        A method for the derivative of the LeakyReLU function.

        This method implements the derivative of the LeakyReLU function using
        the np.where function: 1 if x > 0 else alpha

        Args:
            x: A numpy array of any shape, representing the input to the
            LeakyReLU function.

        Returns:
            A numpy array of the same shape as x, representing the output of
            the derivative of the LeakyReLU function.
        """
        return np.where(x > 0, 1, self.alpha)



import numpy as np
import pytest

from inventronet.activations import Sigmoid


class TestSigmoid:
    # Test the Sigmoid activation function

    @pytest.fixture
    def sigmoid(self) -> Sigmoid:
        # Fixture for the Sigmoid function
        yield Sigmoid()

    @pytest.mark.parametrize(
        "x, y",
        [
            (np.array([0, 1, -1]), np.array([0.5, 0.7310586, 0.2689414])),
            (np.array([2, -2, 0.5]), np.array([0.8807971, 0.1192029, 0.6224593])),
        ],
    )
    def test_sigmoid_function(self, sigmoid: Sigmoid, x, y):
        # Test the function method of the Sigmoid function with different
        # inputs and outputs
        assert np.allclose(sigmoid.function(x), y, atol=1e-4)



import re
from collections import Counter

import numpy as np


class CountVectorizer:
    def __init__(self, lowercase=True, token_pattern=r'\b\w+\b'):
        self.lowercase = lowercase
        self.token_pattern = token_pattern
        self.vocabulary_ = None

    def fit(self, documents):
        word_counts = Counter()
        for doc in documents:
            if self.lowercase:
                doc = doc.lower()
            tokens = re.findall(self.token_pattern, doc)
            word_counts.update(tokens)

        self.vocabulary_ = {word: idx for idx, word in enumerate(word_counts.keys())}

    def transform(self, documents):
        if self.vocabulary_ is None:
            raise RuntimeError("You need to call fit() before calling transform().")

        matrix = []
        for doc in documents:
            if self.lowercase:
                doc = doc.lower()
            tokens = re.findall(self.token_pattern, doc)
            row = [tokens.count(word) for word in self.vocabulary_.keys()]
            matrix.append(row)

        return np.array(matrix)

    def fit_transform(self, documents):
        self.fit(documents)
        return self.transform(documents)



import numpy as np
import pytest
from hypothesis import given, strategies as st, settings

from inventronet.layers import Dropout
from inventronet.layers.shape_error import ShapeError
from tests.layers.conftest import generate_random_array


class TestDropout:
    @pytest.fixture
    def training(self) -> bool:
        yield True

    @pytest.fixture
    def dropout_rate(self) -> float:
        yield 0.2

    @pytest.fixture
    def dropout(self, dropout_rate: float, input_dim: int) -> Dropout:
        yield Dropout(dropout_rate=dropout_rate, input_dim=input_dim)

    @given(
        st.floats(min_value=0.0, max_value=1.0), st.integers(min_value=1, max_value=100)
    )
    def test_dropout_init(self, dropout_rate: float, input_dim: int):
        dropout = Dropout(dropout_rate=dropout_rate, input_dim=input_dim)
        assert dropout.dropout_rate == dropout_rate
        assert dropout.mask is None

    @given(
        st.floats(min_value=0.0, max_value=1.0),
        st.integers(min_value=1, max_value=100),
        st.booleans(),
    )
    @settings(max_examples=20)
    def test_forward(self, dropout_rate: float, input_dim: int, training: bool):
        input_shape = (10, input_dim)
        random_input = generate_random_array(input_shape)

        dropout = Dropout(dropout_rate=dropout_rate, input_dim=input_dim)
        output_array = dropout.forward(random_input, training=training)
        assert output_array.shape == random_input.shape

        if training:
            assert dropout.mask.shape == random_input.shape
            if dropout_rate < 1:
                assert np.allclose(
                    output_array, random_input * dropout.mask / (1 - dropout_rate)
                )

    @given(
        st.floats(min_value=0.0, max_value=1.0),
        st.integers(min_value=1, max_value=100),
        st.booleans(),
    )
    @settings(max_examples=20)
    def test_backward(self, dropout_rate: float, input_dim: int, training: bool):
        input_shape = (10, input_dim)
        random_input = generate_random_array(input_shape)

        dropout = Dropout(dropout_rate=dropout_rate, input_dim=input_dim)
        output_array = dropout.forward(random_input, training=training)
        error = generate_random_array(output_array.shape)
        gradient_array = dropout.backward(error, training=training)

        assert gradient_array.shape == random_input.shape

        if training and dropout_rate < 1:
            assert np.allclose(
                gradient_array, error * dropout.mask / (1 - dropout_rate)
            )

    @given(
        st.floats(min_value=0.0, max_value=1.0),
        st.integers(min_value=1, max_value=100),
        st.booleans(),
    )
    @settings(max_examples=20)
    def test_input_validation(
        self, dropout_rate: float, input_dim: int, training: bool
    ):
        dropout = Dropout(dropout_rate=dropout_rate, input_dim=input_dim)

        with pytest.raises(ValueError):
            dropout.forward(42, training=training)

        with pytest.raises(ValueError):
            dropout.forward([1, 2, 3], training=training)

        with pytest.raises(ValueError):
            dropout.forward("Hello, world!", training=training)

    @given(
        st.floats(min_value=0.0, max_value=1.0),
        st.integers(min_value=1, max_value=100),
        st.booleans(),
    )
    @settings(max_examples=20)
    def test_error_validation(
        self, dropout_rate: float, input_dim: int, training: bool
    ):
        random_input = generate_random_array((5, input_dim))

        dropout = Dropout(dropout_rate=dropout_rate, input_dim=input_dim)
        dropout.forward(random_input, training=True)

        with pytest.raises(ShapeError):
            invalid_error = generate_random_array((2, 3, 4))
            dropout.backward(invalid_error, training=training)

    class TestDroputInit:
        def test_droput_rate(self, dropout: Dropout, dropout_rate: float):
            assert dropout.dropout_rate == dropout_rate

        def test_droput_mask(self, dropout: Dropout):
            assert dropout.mask is None

    class TestForward:
        def test_output_shape(
            self, dropout: Dropout, random_input: np.ndarray, training: bool
        ):
            output_array = dropout.forward(random_input, training=training)
            assert output_array.shape == random_input.shape

        def test_mask_shape(
            self, dropout: Dropout, random_input: np.ndarray, training: bool
        ):
            dropout.forward(random_input, training)
            assert dropout.mask.shape == random_input.shape

        def test_output_value(
            self,
            dropout: Dropout,
            dropout_rate: float,
            random_input: np.ndarray,
            training: bool,
        ):
            output_array = dropout.forward(random_input, training=training)
            assert np.allclose(
                output_array,
                random_input * dropout.mask / (1 - dropout_rate),
            )

    class TestBackward:
        def test_gradient_shape(
            self, dropout: Dropout, random_input: np.ndarray, training: bool
        ):
            output_array = dropout.forward(random_input, training=training)
            error = np.random.randn(*output_array.shape)
            gradient_array = dropout.backward(error, training=training)
            assert gradient_array.shape == random_input.shape

        def test_gradient_value(
            self,
            dropout: Dropout,
            dropout_rate: float,
            random_input: np.ndarray,
            training: bool,
        ):
            output_array = dropout.forward(random_input, training=training)
            error = np.random.randn(*output_array.shape)
            gradient_array = dropout.backward(error, training=training)
            assert np.allclose(
                gradient_array, error * dropout.mask / (1 - dropout_rate)
            )

    @pytest.mark.parametrize("dropout_rate", [0])
    @pytest.mark.parametrize("training", [True])
    def test_dropout_rate_zero_training_true(
        self,
        dropout_rate: float,
        random_input: np.ndarray,
        training: bool,
        input_dim: int,
    ):
        dropout = Dropout(dropout_rate, input_dim=input_dim)
        output_array = dropout.forward(random_input, training=training)
        assert np.allclose(output_array, random_input) and np.all(dropout.mask == 1)

    @pytest.mark.parametrize("dropout_rate", [1])
    @pytest.mark.parametrize("training", [True])
    def test_dropout_rate_one_training_true(
        self,
        dropout_rate: float,
        random_input: np.ndarray,
        training: bool,
        input_dim: int,
    ):
        dropout = Dropout(dropout_rate, input_dim=input_dim)
        output_array = dropout.forward(random_input, training=training)
        print("Output array: ", output_array)
        print("Dropout Mask: ", dropout.mask)
        assert np.all(output_array == 0) and np.all(dropout.mask == 0)

    @pytest.mark.parametrize("dropout_rate", [0])
    @pytest.mark.parametrize("training", [False])
    def test_dropout_rate_zero_training_false(
        self,
        dropout_rate: float,
        random_input: np.ndarray,
        training: bool,
        input_dim: int,
    ):
        dropout = Dropout(dropout_rate, input_dim=input_dim)
        output_array = dropout.forward(random_input, training=training)
        assert np.allclose(output_array, random_input)

    @pytest.mark.parametrize("dropout_rate", [1])
    @pytest.mark.parametrize("training", [False])
    def test_dropout_rate_one_training_false(
        self,
        dropout_rate: float,
        random_input: np.ndarray,
        training: bool,
        input_dim: int,
    ):
        dropout = Dropout(dropout_rate, input_dim=input_dim)
        output_array = dropout.forward(random_input, training=training)
        assert np.allclose(output_array, random_input)



from typing import Type

import numpy as np
import pytest

from inventronet.layers import BatchNormalization
from inventronet.optimizers.optimizer import Optimizer


class TestBatchNormalization:
    @pytest.fixture
    def epsilon(self) -> float:
        yield 1e-5

    @pytest.fixture
    def batch_normalization(self, input_dim: int, epsilon: float) -> BatchNormalization:
        yield BatchNormalization(
            input_dim=input_dim,
            output_dim=input_dim,
            epsilon=epsilon,
        )

    @pytest.fixture
    def forward_output_train(
        self, batch_normalization: BatchNormalization, random_input: np.ndarray
    ) -> np.ndarray:
        yield batch_normalization.forward(random_input, training=True)

    @pytest.fixture
    def forward_output_inference(
        self, batch_normalization: BatchNormalization, random_input: np.ndarray
    ) -> np.ndarray:
        yield batch_normalization.forward(random_input, training=False)

    def test_mean_and_variance_normalization(
        self, batch_normalization: BatchNormalization, random_input: np.ndarray
    ):
        output = batch_normalization.forward(random_input, training=True)
        assert np.isclose(np.mean(output, axis=0), 0, atol=1e-6).all()
        assert np.isclose(np.var(output, axis=0), 1, atol=1e-4).all()

    def test_running_mean_and_variance_update(
        self, batch_normalization: BatchNormalization, random_input: np.ndarray
    ):
        initial_running_mean = batch_normalization.running_mean.copy()
        initial_running_var = batch_normalization.running_var.copy()
        batch_normalization.forward(random_input, training=True)
        assert not np.allclose(batch_normalization.running_mean, initial_running_mean)
        assert not np.allclose(batch_normalization.running_var, initial_running_var)

    def test_use_running_statistics_during_inference(
        self, batch_normalization: BatchNormalization, random_input: np.ndarray
    ):
        batch_normalization.forward(random_input, training=True)
        output = batch_normalization.forward(random_input, training=False)
        expected_output = (
            batch_normalization.gamma
            * (random_input - batch_normalization.running_mean)
            / np.sqrt(batch_normalization.running_var + batch_normalization.epsilon)
            + batch_normalization.beta
        )
        assert np.allclose(output, expected_output)

    def test_differentiability(
        self,
        batch_normalization: BatchNormalization,
        random_input: np.ndarray,
        random_error: np.ndarray,
    ):
        batch_normalization.forward(random_input, training=True)
        try:
            batch_normalization.backward(random_error, 0.01)
        except Exception as e:
            pytest.fail(f"Differentiability test failed with exception: {e}")

    def test_parameter_updates(
        self,
        batch_normalization: BatchNormalization,
        random_input: np.ndarray,
        random_error: np.ndarray,
        optimizer: Type[Optimizer],
    ):
        original_gamma = batch_normalization.gamma.copy()
        original_beta = batch_normalization.beta.copy()

        batch_normalization.forward(random_input, training=True)
        batch_normalization.backward(random_error)

        optimizer.update(
            0, batch_normalization.parameters, batch_normalization.gradients
        )

        assert not np.allclose(batch_normalization.gamma, original_gamma, atol=1e-6)
        assert not np.allclose(batch_normalization.beta, original_beta, atol=1e-6)

    def test_output_dimension(
        self, batch_normalization: BatchNormalization, random_input: np.ndarray
    ):
        output = batch_normalization.forward(random_input, training=True)
        assert output.shape == random_input.shape

    class TestForward:
        def test_forward_training_mode_without_activation(
            self,
            random_input: np.ndarray,
            batch_normalization: BatchNormalization,
            forward_output_train: np.ndarray,
        ):
            batch_mean = np.mean(random_input, axis=0)
            batch_var = np.var(random_input, axis=0)
            normalized_input = (random_input - batch_mean) / np.sqrt(
                batch_var + batch_normalization.epsilon
            )
            expected_output = (
                batch_normalization.gamma * normalized_input + batch_normalization.beta
            )
            assert np.allclose(forward_output_train, expected_output)

        def test_forward_inference_mode_without_activation(
            self,
            batch_normalization: BatchNormalization,
            random_input: np.ndarray,
            forward_output_inference: np.ndarray,
        ):
            normalized_input = (
                random_input - batch_normalization.running_mean
            ) / np.sqrt(batch_normalization.running_var + batch_normalization.epsilon)
            expected_output = (
                batch_normalization.gamma * normalized_input + batch_normalization.beta
            )
            assert np.allclose(forward_output_inference, expected_output)

    class TestBackward:
        @pytest.fixture
        def batch_size(self, random_input: np.ndarray) -> int:
            yield random_input.shape[0]

        @pytest.fixture
        def gradients(self, random_input: np.ndarray) -> np.ndarray:
            yield np.random.randn(*random_input.shape)

        def test_backward_pass(
            self,
            batch_normalization: BatchNormalization,
            gradients: np.ndarray,
            random_input: np.ndarray,
            epsilon: float,
        ) -> None:
            batch_mean = np.mean(random_input, axis=0)
            batch_var = np.var(random_input, axis=0)
            normalized_input = (random_input - batch_mean) / np.sqrt(
                batch_var + batch_normalization.epsilon
            )
            expected_output = (
                gradients
                - np.mean(gradients, axis=0)
                - normalized_input * np.mean(gradients * normalized_input, axis=0)
            ) / np.sqrt(batch_var + epsilon)
            batch_normalization.forward(random_input, training=True)
            output = batch_normalization.backward(gradients)
            assert np.allclose(output, expected_output, rtol=0.1)

        def test_backward_gamma_without_activation(
            self,
            batch_normalization: BatchNormalization,
            random_input: np.ndarray,
            gradients: np.ndarray,
        ):
            batch_normalization.forward(random_input, training=True)
            batch_normalization.backward(gradients)

            batch_mean = np.mean(random_input, axis=0)
            batch_var = np.var(random_input, axis=0)
            normalized_input = (random_input - batch_mean) / np.sqrt(
                batch_var + batch_normalization.epsilon
            )
            expected_gradient_gamma = np.sum(gradients * normalized_input, axis=0)
            assert np.allclose(
                batch_normalization.gradients["gamma"], expected_gradient_gamma
            )

        def test_backward_beta_without_activation(
            self,
            batch_normalization: BatchNormalization,
            random_input: np.ndarray,
            gradients: np.ndarray,
        ):
            batch_normalization.forward(random_input, training=True)
            batch_normalization.backward(gradients)

            expected_gradient_beta = np.sum(gradients, axis=0)
            assert np.allclose(
                batch_normalization.gradients["beta"], expected_gradient_beta
            )



from typing import Optional
import numpy as np
from .activation import Activation


class Linear(Activation):
    """
    A class for the linear activation function.

    This class implements the linear activation function and its derivative,
    which are both simply the identity function.
    """

    def function(self, x: np.ndarray) -> np.ndarray:
        """
        The linear activation function.

        This method implements the linear activation function: `x`.

        Args:
            x: A numpy array of any shape, representing the input to the
            activation function.

        Returns:
            A numpy array of the same shape as x, representing the output of
            the activation function.
        """
        return x

    def derivative(
        self, x: np.ndarray, output: Optional[np.ndarray] = None
    ) -> np.ndarray:
        """
        The derivative of the linear activation function.

        This method implements the derivative of the linear activation
        function: `1`.

        Args:
            x: A numpy array of any shape, representing the input to
            the activation function.

        Returns:
            A numpy array of the same shape as x, representing the output
            of the derivative of the activation function.
        """
        return np.ones_like(x)



from typing import Dict

import numpy as np
import pytest


@pytest.fixture
def params() -> Dict[str, np.ndarray]:
    yield {
        "weights": np.array([[1.0, 2.0], [3.0, 4.0]]),
        "biases": np.array([0.5, 0.5]),
    }


@pytest.fixture
def gradients() -> Dict[str, np.ndarray]:
    yield {
        "weights": np.array([[0.1, 0.2], [0.3, 0.4]]),
        "biases": np.array([0.1, -0.1]),
    }


@pytest.fixture
def zero_params() -> Dict[str, np.ndarray]:
    yield {
        "weights": np.zeros((2, 2)),
        "biases": np.zeros(2),
    }


@pytest.fixture
def zero_gradients() -> Dict[str, np.ndarray]:
    yield {
        "weights": np.zeros((2, 2)),
        "biases": np.zeros(2),
    }



class ShapeError(Exception):
    """An exception class for shape mismatch."""

    def __init__(self, message: str):
        """Initialize the exception with a message."""
        super().__init__(message)



from .binary_cross_entropy import BinaryCrossEntropy
from .categorical_cross_entropy import CategoricalCrossEntropy
from .mean_absolute_error import MeanAbsoluteError
from .mean_squared_error import MeanSquaredError

__all__ = [
    "BinaryCrossEntropy",
    "CategoricalCrossEntropy",
    "MeanAbsoluteError",
    "MeanSquaredError",
]



import numpy as np
import pytest

from inventronet.metrics import Accuracy


@pytest.fixture
def accuracy() -> Accuracy:
    yield Accuracy()


@pytest.mark.parametrize(
    "y_true, y_pred, expected",
    [
        (
            np.array([1, 0, 0, 1, 1]),
            np.array([0.9, 0.1, 0.2, 0.8, 0.7]),
            1.0,
        ),  # exact match
        (
            np.array([1, 0, 0, 1, 1]),
            np.array([0.6, 0.4, 0.3, 0.7, 0.5]),
            0.8,
        ),  # small difference
        (
            np.array([1, 0, 0, 1, 1]),
            np.array([0.1, 0.9, 0.8, 0.2, 0.3]),
            0.0,
        ),  # large difference
        (
            [1, 0, 0, 1, 1],
            [0.9, 0.1, 0.2, 0.8, 0.7],
            1.0,
        ),  # Python list input
        (
            np.array([1, 0, 1, 0, 1]),
            np.array([0.6, 0.4, 0.6, 0.4, 0.6]),
            1.0,  # Update the expected value to 1.0
        ),  # equal number of correct and incorrect predictions
    ],
)
def test_accuracy_call(y_true, y_pred, expected, accuracy: Accuracy):
    np.testing.assert_almost_equal(accuracy.call(y_true, y_pred), expected, decimal=7)



import numpy as np
import pytest

from inventronet.losses import CategoricalCrossEntropy


class TestCCE:
    @pytest.fixture
    def cce(self) -> CategoricalCrossEntropy:
        yield CategoricalCrossEntropy()

    @pytest.mark.parametrize(
        "y_true, y_pred, expected",
        [
            (
                    np.array([[1, 0], [0, 1], [1, 0]]),
                    np.array([[0.9, 0.1], [0.1, 0.9], [0.9, 0.1]]),
                    0.10536052,
            ),  # high confidence
            (
                    np.array([[1, 0], [0, 1], [1, 0]]),
                    np.array([[0.6, 0.4], [0.4, 0.6], [0.6, 0.4]]),
                    0.51082562,
            ),  # medium confidence
            (
                    np.array([[1, 0], [0, 1], [1, 0]]),
                    np.array([[0.5, 0.5], [0.5, 0.5], [0.5, 0.5]]),
                    0.69314718,
            ),  # low confidence
        ],
    )
    def test_cce_function(self, y_true, y_pred, expected, cce: CategoricalCrossEntropy):
        # Compare the expected and actual outputs with a precision of
        # 7 decimal places
        np.testing.assert_almost_equal(
            cce.function(y_true, y_pred), expected, decimal=7
        )

    @pytest.mark.parametrize(
        "y_true, y_pred, expected",
        [
            (
                    np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
                    np.array([[0.9, 0.05, 0.05], [0.05, 0.9, 0.05], [0.05, 0.05, 0.9]]),
                    np.array([[-0.1, 0.05, 0.05], [0.05, -0.1, 0.05], [0.05, 0.05, -0.1]]),
            ),  # high confidence
            (
                    np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
                    np.array([[0.6, 0.2, 0.2], [0.2, 0.6, 0.2], [0.2, 0.2, 0.6]]),
                    np.array([[-0.4, 0.2, 0.2], [0.2, -0.4, 0.2], [0.2, 0.2, -0.4]]),
            ),  # medium confidence
            (
                    np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
                    np.array([[0.33, 0.33, 0.33], [0.33, 0.33, 0.33], [0.33, 0.33, 0.33]]),
                    np.array(
                        [[-0.67, 0.33, 0.33], [0.33, -0.67, 0.33], [0.33, 0.33, -0.67]]
                    ),
            ),  # low confidence
        ],
    )
    def test_cce_derivative(
            self, y_true, y_pred, expected, cce: CategoricalCrossEntropy
    ):
        # Compare the expected and actual outputs with a precision of 7
        # decimal places
        np.testing.assert_almost_equal(
            cce.gradient(y_true, y_pred), expected, decimal=7
        )

    def test_cce_non_negativity(self, cce: CategoricalCrossEntropy):
        """
        Test that the CCE loss is non-negative.
        """
        y_true = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        y_pred = np.random.rand(3, 3)
        y_pred /= y_pred.sum(axis=1, keepdims=True)

        loss = cce(y_true, y_pred)
        assert loss >= 0, f"CCE loss should be non-negative, got {loss}"

    def test_cce_lower_bound(self, cce: CategoricalCrossEntropy):
        """
        Test that the CCE loss is 0 when the true and predicted labels are
        the same.
        """
        y_true = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        y_pred = y_true.copy()

        loss = cce(y_true, y_pred)
        assert np.isclose(loss, 0), f"CCE loss should be 0, got {loss}"

    def test_cce_gradient_property(self, cce: CategoricalCrossEntropy):
        """
        Test that the gradient of the CCE loss is 0 when the true and
        predicted labels are the same.
        """
        y_true = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        y_pred = y_true.copy()

        gradient = cce.gradient(y_true, y_pred)
        assert np.allclose(gradient, 0), f"Gradient should be 0, got {gradient}"

    def test_non_negative(self, cce: CategoricalCrossEntropy) -> None:
        """
        Test that the categorical cross entropy loss is non-negative for any
          true and predicted labels.
        """
        # Generate some random true and predicted labels
        y_true = np.random.randint(0, 3, size=(10, 1))
        y_pred = np.random.rand(10, 3)
        # Compute the loss
        loss = cce.function(y_true, y_pred)
        # Assert that the loss is non-negative
        assert loss >= 0



import numpy as np


def train_test_split(X, y, test_size=0.2, random_state=None):
    if random_state is not None:
        np.random.seed(random_state)

    data_size = len(X)
    indices = np.arange(data_size)
    np.random.shuffle(indices)

    test_count = int(test_size * data_size)
    train_indices = indices[:-test_count]
    test_indices = indices[-test_count:]

    X_train = X[train_indices]
    X_test = X[test_indices]
    y_train = y[train_indices]
    y_test = y[test_indices]

    return X_train, X_test, y_train, y_test



import numpy as np
from .optimizer import Optimizer


class Adam(Optimizer):
    def __init__(
        self,
        learning_rate: float = 0.001,
        beta1: float = 0.9,
        beta2: float = 0.999,
        epsilon: float = 1e-8,
    ):
        self.learning_rate = learning_rate
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon
        self.m = {}  # Initialize m as an empty dictionary
        self.v = {}  # Initialize v as an empty dictionary
        self.t = 0

    def update(self, layer_id: int, params: dict, gradients: dict):
        if layer_id not in self.m:
            self.m[layer_id] = {key: np.zeros_like(value) for key, value in params.items()}
            self.v[layer_id] = {key: np.zeros_like(value) for key, value in params.items()}

        self.t += 1

        for key in params.keys():
            self.m[layer_id][key] = self.beta1 * self.m[layer_id][key] + (1 - self.beta1) * gradients[key]
            self.v[layer_id][key] = self.beta2 * self.v[layer_id][key] + (1 - self.beta2) * np.square(gradients[key])

            m_hat = self.m[layer_id][key] / (1 - self.beta1 ** self.t)
            v_hat = self.v[layer_id][key] / (1 - self.beta2 ** self.t)

            params[key] -= self.learning_rate * m_hat / (np.sqrt(v_hat) + self.epsilon)



import numpy as np

from .layer import Layer
from .shape_error import ShapeError


class Dropout(Layer):
    def __init__(self, dropout_rate: float, input_dim: int) -> None:
        """Initialize the dropout layer with the given dropout rate.

        Args:
            dropout_rate: The probability of dropping out each neuron.
        """
        super().__init__(input_dim=input_dim, output_dim=input_dim, activation=None)
        self.dropout_rate: float = dropout_rate
        self.mask: np.ndarray = None

    def forward(self, inputs: np.ndarray, training: bool = False) -> np.ndarray:
        """Perform the layer operation on the inputs.

        Args:
            training: A boolean indicating whether the layer is in training
            inputs: The input array of any shape.

        Returns:
            The output array of the same shape as the input.
        """
        self.previous_layer_output: np.ndarray = inputs

        if not isinstance(inputs, np.ndarray):
            raise ValueError("The input must be a numpy array.")

        if training:
            self.mask = np.random.rand(*inputs.shape) >= self.dropout_rate
            output = inputs * self.mask
            if self.dropout_rate != 1:
                output = output / (1 - self.dropout_rate)
            return output
        else:
            return inputs

    def backward(
        self,
        error: np.ndarray,
        prev_output: np.ndarray = None,
        training: bool = False,
    ) -> np.ndarray:
        """Perform the backpropagation on the error.

        Args:
            training:
            prev_output:
            error: The error array of the same shape as the output.

        Returns:
            The output error of the same shape as the input.
        """
        if not isinstance(error, np.ndarray):
            raise ValueError("The error must be a numpy array.")

        if error.shape != self.previous_layer_output.shape:
            raise ShapeError(
                f"The shape of the error {error.shape} does not match the "
                + "shape of the input {self.previous_layer_output.shape}."
            )

        if training:
            output_error = error * self.mask
            if self.dropout_rate != 1:
                output_error = output_error / (1 - self.dropout_rate)
            return output_error
        else:
            return error

    def get_parameters(self):
        return []



import numpy as np
from .loss import Loss


class BinaryCrossEntropy(Loss):
    def function(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        """
        Computes the binary cross-entropy loss between the true labels and the
            predicted values.

        Parameters
        ----------
        y_true : np.ndarray
            The true labels of the data, encoded as one-hot vectors or as
                scalars.
            Shape: (n_samples, n_classes) or (n_samples,)
        y_pred : np.ndarray
            The predicted values of the data, output of the sigmoid function.
            Shape: (n_samples, n_classes) or (n_samples,)

        Returns
        -------
        float
            The loss, averaged over all samples and classes or samples.
        """
        # Clip the predictions to avoid log(0) or log(1) errors
        y_pred = np.clip(y_pred, 1e-15, 1 - 1e-15)
        # Compute the binary cross-entropy loss
        loss = -np.mean(y_true * np.log(y_pred) + (1 - y_true) * np.log(1 - y_pred))
        return loss

    def gradient(self, y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
        """
        Computes the derivative of the binary cross-entropy loss with respect
            to the predictions.

        Parameters
        ----------
        y_true : np.ndarray
            The true labels of the data, encoded as one-hot vectors or as
                scalars.
            Shape: (n_samples, n_classes) or (n_samples,)
        y_pred : np.ndarray
            The predicted values of the data, output of the sigmoid function.
            Shape: (n_samples, n_classes) or (n_samples,)

        Returns
        -------
        np.ndarray
            The derivative of the loss, scaled by the number of samples.
            Shape: (n_samples, n_classes) or (n_samples,)
        """
        # Clip the predictions to avoid division by 0 errors
        y_pred = np.clip(y_pred, 1e-15, 1 - 1e-15)
        # Compute the derivative of the binary cross-entropy loss
        grad = -(y_true / y_pred - (1 - y_true) / (1 - y_pred)) / len(y_true)
        return grad



import numpy as np

from .loss import Loss


class MeanAbsoluteError(Loss):
    def function(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        """
        Computes the mean absolute error loss between the true labels and the
        predicted values.

        Parameters
        ----------
        y_true : np.ndarray
            The true labels of the data, encoded as scalars.
            Shape: (n_samples,)
        y_pred : np.ndarray
            The predicted values of the data, output of the activation
            function.
            Shape: (n_samples,)

        Returns
        -------
        float
            The loss, averaged over all samples.
        """
        # Compute the absolute error
        error = np.abs(y_true - y_pred)
        # Compute the mean absolute error
        loss = np.mean(error)
        return loss

    def gradient(self, y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
        """
        Computes the derivative of the mean absolute error loss with respect to
        the predictions.

        Parameters
        ----------
        y_true : np.ndarray
            The true labels of the data, encoded as scalars.
            Shape: (n_samples,)
        y_pred : np.ndarray
            The predicted values of the data, output of the
            activation function.
            Shape: (n_samples,)

        Returns
        -------
        np.ndarray
            The derivative of the loss, scaled by the number of samples.
            Shape: (n_samples,)
        """
        # Compute the sign of the error
        sign = np.sign(y_true - y_pred)
        # Compute the derivative of the mean absolute error
        derivative = -sign / y_true.size
        print(f"Gradient: {derivative}")
        return derivative



from abc import ABC, abstractmethod


class Optimizer(ABC):
    @abstractmethod
    def update(self, layer_id: int, params: dict, gradients: dict):
        raise NotImplementedError("The update method must be implemented.")



from .leaky_relu import LeakyReLU
from .linear import Linear
from .relu import ReLU
from .sigmoid import Sigmoid
from .softmax import SoftMax
from .tanh import Tanh

__all__ = ["LeakyReLU", "Linear", "ReLU", "Sigmoid", "SoftMax", "Tanh"]



from abc import ABC, abstractmethod
from typing import Optional
import numpy as np


class Activation(ABC):
    """
    An abstract class for an activation function.

    An activation function is a function that maps an input to an output,
    usually in the range (0, 1) or (-1, 1).
    It is used to introduce non-linearity to a neural network and to control
    the output of a neuron.

    Attributes:
        None
    """

    @abstractmethod
    def function(self, x: np.ndarray) -> np.ndarray:
        """
        An abstract method for the function.

        This method should take an input x and return the output of the
        activation function.

        Args:
            x: A numpy array of shape (n,) or (m, n), representing the input
            to the activation function.
            output: Optional precomputed output of the activation function
            to avoid redundant computation in the derivative calculation.

        Returns:
            A numpy array of the same shape as x, representing the output of
            the activation function.
        """
        raise NotImplementedError("Activation.function is not implemented.")

    @abstractmethod
    def derivative(
        self, x: np.ndarray, output: Optional[np.ndarray] = None
    ) -> np.ndarray:
        """
        An abstract method for the derivative.

        This method should take an input x and return the output of the
        derivative of the activation function.

        Args:
            x: A numpy array of shape (n,) or (m, n), representing the input
            to the activation function.

        Returns:
            A numpy array of shape (n, n) or (m, n, n), representing the
            output of the derivative of the activation function.
        """
        raise NotImplementedError("Activation.derivative is not implemented.")

    def __call__(self, x: np.ndarray) -> np.ndarray:
        """
        A method to make the Activation object callable like a function.

        This method overrides the __call__ method of the Object class, and
        returns the output of the function method.

        Args:
            x: A numpy array of any shape, representing the input to the
            activation function.

        Returns:
            A numpy array of the same shape as x, representing the output
            of the activation function.
        """
        return self.function(x)



import numpy as np
from .loss import Loss


class CategoricalCrossEntropy(Loss):
    def function(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        """
        Computes the categorical cross-entropy loss between the true labels
            and the predicted values.

        Parameters
        ----------
        y_true : np.ndarray
            The true labels of the data, encoded as one-hot vectors.
            Shape: (n_samples, n_classes)
        y_pred : np.ndarray
            The predicted values of the data, output of the softmax function.
            Shape: (n_samples, n_classes)

        Returns
        -------
        float
            The loss, averaged over all samples and classes.
        """
        # Clip the predictions to avoid log(0) errors
        y_pred = np.clip(y_pred, 1e-15, 1 - 1e-15)
        # Compute the categorical cross-entropy loss
        loss = -np.sum(y_true * np.log(y_pred)) / len(y_true)
        return loss

    def gradient(self, y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
        """
        Computes the derivative of the categorical cross-entropy loss with
            respect to the predictions.

        Parameters
        ----------
        y_true : np.ndarray
            The true labels of the data, encoded as one-hot vectors.
            Shape: (n_samples, n_classes)
        y_pred : np.ndarray
            The predicted values of the data, output of the softmax function.
            Shape: (n_samples, n_classes)

        Returns
        -------
        np.ndarray
            The derivative of the loss, scaled by the number of samples.
            Shape: (n_samples, n_classes)
        """
        # Clip the predictions to avoid division by zero errors in the
        # gradient calculation
        y_pred = np.clip(y_pred, 1e-15, 1 - 1e-15)

        # Compute the derivative of the categorical cross-entropy loss
        grad = -(y_true - y_pred)
        return grad



import inspect
from typing import Type

import numpy as np
import pytest
from _pytest.fixtures import SubRequest

import inventronet.losses as losses_module
import inventronet.metrics as metrics_module
import inventronet.optimizers as optimizers_module
from inventronet.layers import Dense
from inventronet.losses.loss import Loss
from inventronet.metrics.metric import Metric
from inventronet.models import Sequential
from inventronet.optimizers.optimizer import Optimizer

# Get all classes in the metrics module that are subclasses of Metric
all_metrics = [
    metric
    for _, metric in inspect.getmembers(metrics_module, inspect.isclass)
    if issubclass(metric, Metric) and metric != Metric
]

# Get all classes in the losses module that are subclasses of Loss
all_losses = [
    loss
    for _, loss in inspect.getmembers(losses_module, inspect.isclass)
    if issubclass(loss, Loss) and loss != Loss
]

all_optimizers = [
    optimizer
    for _, optimizer in inspect.getmembers(optimizers_module, inspect.isclass)
    if issubclass(optimizer, Optimizer) and optimizer != Optimizer
]


@pytest.fixture(params=all_metrics)
def metric(request: SubRequest) -> Type[Metric]:
    yield request.param()


@pytest.fixture(params=all_losses)
def loss(request: SubRequest) -> Type[Loss]:
    yield request.param()


@pytest.fixture(params=all_optimizers)
def optimizer(request: SubRequest) -> Type[Optimizer]:
    opt: Type[Optimizer] = request.param()
    # Increase the learning rate
    opt.learning_rate = 0.1
    yield opt


@pytest.fixture
def dummy_sequential_model() -> Sequential:
    """Create a dummy sequential model with two layers"""
    # Create a sequential model
    model = Sequential()
    # Add a dense layer with input dimension 3 and output dimension 2
    model.add(Dense(3, 2))
    # Add a dense layer with input dimension 2 and output dimension 1
    model.add(Dense(2, 1))
    # Return the model
    return model


@pytest.fixture
def dummy_compiled_sequential_model(
        dummy_sequential_model: Sequential,
        loss: Type[Loss],
        optimizer: Type[Optimizer],
        metric: Type[Metric],
) -> Sequential:
    """Compile the sequential model with a loss function, and a metric."""
    dummy_sequential_model.compile(loss, optimizer, [metric])
    # Return the model
    return dummy_sequential_model


@pytest.fixture
def x() -> np.ndarray:
    # Create some dummy input data
    yield np.array([[0, 0, 1], [0, 1, 0], [1, 0, 0], [1, 1, 0]])


@pytest.fixture
def y() -> np.ndarray:
    # Create some dummy output data
    yield np.array([[0], [0.5], [0.5], [1]])


def test_add(dummy_sequential_model: Sequential):
    """Test the add method of the sequential model."""
    # Try to add a layer with input dimension 4 and output dimension 2
    with pytest.raises(AssertionError):
        dummy_sequential_model.add(Dense(4, 2))
    # Check that the model still has two layers
    assert len(dummy_sequential_model.layers) == 2


class TestCompile:

    def test_model_loss(
            self,
            dummy_sequential_model: Sequential,
            loss: Type[Loss],
            optimizer: Type[Optimizer],
            metric: Type[Metric],
    ):
        # Compile the model with a loss function and a metric
        dummy_sequential_model.compile(loss, optimizer, [metric])
        # Check that the model has the loss function
        assert dummy_sequential_model.loss == loss

    def test_model_metric_before(self, dummy_sequential_model: Sequential):
        with pytest.raises(AttributeError):
            dummy_sequential_model.metric

    def test_model_metric(
            self,
            dummy_sequential_model: Sequential,
            loss: Type[Loss],
            optimizer: Type[Optimizer],
            metric: Type[Metric],
    ):
        # Compile the model with a loss function and a metric
        dummy_sequential_model.compile(loss, optimizer, [metric])
        # Check that the model has the metrics
        assert dummy_sequential_model.metrics == [metric]


class TestFit:
    def test_fit_history_is_dict(
            self,
            dummy_compiled_sequential_model: Sequential,
            x: np.ndarray,
            y: np.ndarray,
    ):
        dummy_compiled_sequential_model.fit(x, y, 5)
        history = dummy_compiled_sequential_model.history
        assert isinstance(history, dict)

    def test_fit_history_has_loss_and_metric(
            self,
            dummy_compiled_sequential_model: Sequential,
            x: np.ndarray,
            y: np.ndarray,
    ):
        dummy_compiled_sequential_model.fit(x, y, 5)
        history = dummy_compiled_sequential_model.history
        metric_class_name = dummy_compiled_sequential_model.metrics[0].__class__.__name__
        assert "loss" in history and metric_class_name in history.keys()

    def test_fit_history_correct_length(
            self,
            dummy_compiled_sequential_model: Sequential,
            x: np.ndarray,
            y: np.ndarray,
    ):
        dummy_compiled_sequential_model.fit(x, y, 5)
        history = dummy_compiled_sequential_model.history
        assert len(history["loss"]) == 5 and all(len(history[key]) == 5 for key in history.keys() if "metric" in key)

    def test_fit_history_values_are_floats(
            self,
            dummy_compiled_sequential_model: Sequential,
            x: np.ndarray,
            y: np.ndarray,
    ):
        dummy_compiled_sequential_model.fit(x, y, 5)
        history = dummy_compiled_sequential_model.history
        assert all(isinstance(value, float) for value in history["loss"]) and all(
            all(isinstance(value, float) for value in history[key]) for key in history.keys() if "metric" in key)

    def test_fit_history_metric_increases(
            self,
            dummy_compiled_sequential_model: Sequential,
            x: np.ndarray,
            y: np.ndarray,
    ):
        dummy_compiled_sequential_model.fit(x, y, 5)
        history = dummy_compiled_sequential_model.history
        for key in history.keys():
            if "metric" in key:
                assert all(
                    history[key][i] <= history[key][i + 1] for i in range(len(history[key]) - 1)
                )

    class TestFitUpdatesWeightsAndBiases:
        def test_layer_1_weights(
                self,
                dummy_compiled_sequential_model: Sequential,
                x: np.ndarray,
                y: np.ndarray,
        ):
            # Get the initial weights and biases of the layers
            w1, _ = dummy_compiled_sequential_model.layers[0].get_parameters()
            print("Initial weights:", w1)

            # Fit the model for one epoch
            dummy_compiled_sequential_model.fit(x, y, 1)

            # Get the updated weights and biases of the layers
            w1_new, _ = dummy_compiled_sequential_model.layers[0].get_parameters()
            print("Updated weights:", w1_new)

            # Check that the weights have changed
            assert not np.array_equal(w1, w1_new)

        def test_layer_1_biases(
                self,
                dummy_compiled_sequential_model: Sequential,
                x: np.ndarray,
                y: np.ndarray,
        ):
            # Get the initial weights and biases of the layers
            _, b1 = dummy_compiled_sequential_model.layers[0].get_parameters()
            # Fit the model for one epoch
            dummy_compiled_sequential_model.fit(x, y, 1)
            # Get the updated weights and biases of the layers
            _, b1_new = dummy_compiled_sequential_model.layers[0].get_parameters()
            # Check that the biases have changed
            assert not np.array_equal(b1, b1_new)

        def test_layer_2_weights(
                self,
                dummy_compiled_sequential_model: Sequential,
                x: np.ndarray,
                y: np.ndarray,
        ):
            # Get the initial weights and biases of the layers
            w2, _ = dummy_compiled_sequential_model.layers[1].get_parameters()
            # Fit the model for one epoch
            dummy_compiled_sequential_model.fit(x, y, 1)
            # Get the updated weights and biases of the layers
            w2_new, _ = dummy_compiled_sequential_model.layers[1].get_parameters()
            # Check that the weights have changed
            assert not np.array_equal(w2, w2_new)

        def test_layer_2_biases(
                self,
                dummy_compiled_sequential_model: Sequential,
                x: np.ndarray,
                y: np.ndarray,
        ):
            # Get the initial weights and biases of the layers
            _, b2 = dummy_compiled_sequential_model.layers[1].get_parameters()
            # Fit the model for one epoch
            dummy_compiled_sequential_model.fit(x, y, 1)
            # Get the updated weights and biases of the layers
            _, b2_new = dummy_compiled_sequential_model.layers[1].get_parameters()
            # Check that the biases have changed
            assert not np.array_equal(b2, b2_new)


class TestSequentialModelProperties:
    def test_layers_property(self, dummy_sequential_model: Sequential):
        assert hasattr(dummy_sequential_model, "layers")
        assert len(dummy_sequential_model.layers) == 2
        assert all(isinstance(layer, Dense) for layer in dummy_sequential_model.layers)

    def test_loss_property(
            self, dummy_compiled_sequential_model: Sequential, loss: Type[Loss]
    ):
        assert hasattr(dummy_compiled_sequential_model, "loss")
        assert dummy_compiled_sequential_model.loss == loss

    def test_optimizer_property(
            self, dummy_compiled_sequential_model: Sequential, optimizer: Type[Optimizer]
    ):
        assert hasattr(dummy_compiled_sequential_model, "optimizer")
        assert dummy_compiled_sequential_model.optimizer == optimizer

    def test_metrics_property(
            self, dummy_compiled_sequential_model: Sequential, metric: Type[Metric]
    ):
        assert hasattr(dummy_compiled_sequential_model, "metrics")
        assert len(dummy_compiled_sequential_model.metrics) == 1
        assert dummy_compiled_sequential_model.metrics == [metric]


def test_layer_compatibility(dummy_sequential_model: Sequential):
    # Try to add a layer with input dimension 4 and output dimension 2
    with pytest.raises(AssertionError):
        dummy_sequential_model.add(Dense(4, 2))


def test_layer_ordering(dummy_sequential_model: Sequential):
    first_layer = dummy_sequential_model.layers[0]
    second_layer = dummy_sequential_model.layers[1]

    assert isinstance(first_layer, Dense)
    assert isinstance(second_layer, Dense)

    assert first_layer.output_dim == 2
    assert second_layer.input_dim == 2


class TestPredict:
    def test_predict_output_shape(
            self, dummy_compiled_sequential_model: Sequential, x: np.ndarray
    ):
        predictions = dummy_compiled_sequential_model.predict(x)
        assert predictions.shape == (x.shape[0], 1)

    def test_predict_output_values(
            self, dummy_compiled_sequential_model: Sequential, x: np.ndarray
    ):
        predictions = dummy_compiled_sequential_model.predict(x)
        assert np.all((predictions >= 0) & (predictions <= 1))


class TestEvaluate:
    def test_evaluate_loss(
            self,
            dummy_compiled_sequential_model: Sequential,
            x: np.ndarray,
            y: np.ndarray,
    ):
        loss_value, _ = dummy_compiled_sequential_model.evaluate(x, y)
        assert isinstance(loss_value, float)

    def test_evaluate_metric(
            self,
            dummy_compiled_sequential_model: Sequential,
            x: np.ndarray,
            y: np.ndarray,
    ):
        _, metric_value = dummy_compiled_sequential_model.evaluate(x, y)
        assert isinstance(metric_value, list)



import numpy as np
import pytest

from inventronet.activations import SoftMax


class TestSoftMax:
    # Test the SoftMax activation function

    @pytest.fixture
    def softmax(self) -> SoftMax:
        # Fixture for the SoftMax function
        yield SoftMax()

    @pytest.mark.parametrize(
        "x, y",
        [
            (
                np.array([0, 1, -1]),
                np.array([0.24472847, 0.66524096, 0.09003057]),
            ),
            (
                np.array([2, -2, 0.5]),
                np.array([0.80551241, 0.01475347, 0.17973411]),
            ),
        ],
    )
    def test_softmax_function(self, softmax: SoftMax, x, y):
        # Test the function method of the SoftMax function with different
        # inputs and outputs
        assert np.allclose(softmax.function(x), y, atol=1e-4)

    @pytest.mark.parametrize(
        "x, y",
        [
            (
                np.array([0, 1, -1]),
                np.array(
                    [
                        [0.18483646, -0.1628034, -0.02203305],
                        [-0.1628034, 0.22269507, -0.05989167],
                        [-0.02203305, -0.05989167, 0.08192472],
                    ]
                ),
            ),
            (
                np.array([2, -2, 0.5]),
                np.array(
                    [
                        [0.15666217, -0.01188411, -0.14477806],
                        [-0.01188411, 0.01453581, -0.0026517],
                        [-0.14477806, -0.0026517, 0.14742976],
                    ]
                ),
            ),
        ],
    )
    def test_softmax_derivative(self, softmax: SoftMax, x, y):
        # Test the derivative method of the SoftMax function with different
        # inputs and outputs
        jacobian = softmax.derivative(x)
        assert np.allclose(jacobian, y, atol=1e-4)



from setuptools import setup, find_packages

package_name = "inventronet"
package_version = "0.0.4-alpha"
package_description = "A package for building and testing neural networks"

# Read the README.md file
with open("README.md", "r", encoding="utf-8") as f:
    package_long_description = f.read()

package_url = "https://github.com/inventrohyder/inventronet"
package_author = "inventrohyder"
package_author_email = "haitham.hyder@uni.minerva.edu"
package_license = "MIT"
package_classifiers = [
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
]
package_keywords = "neural network, machine learning, artificial intelligence"
package_install_requires = [
    "numpy",
]

package_extras_require = {
    "test": ["pytest", "pytest-cov"],
    "docs": ["sphinx", "sphinx-rtd-theme"],
    "lint": ["flake8", "ruff", "black"],
}

setup(
    name=package_name,
    version=package_version,
    description=package_description,
    long_description=package_long_description,
    long_description_content_type="text/markdown",
    url=package_url,
    author=package_author,
    author_email=package_author_email,
    license=package_license,
    packages=find_packages(),
    install_requires=package_install_requires,
    classifiers=package_classifiers,
    extras_require=package_extras_require,
    python_requires=">=3.9",
)



import numpy as np
import pytest

from inventronet.losses import MeanAbsoluteError


class TestMAE:
    # Create an instance of the MAE loss function
    @pytest.fixture
    def mae(self) -> MeanAbsoluteError:
        yield MeanAbsoluteError()

    # Test the function method of the MSE loss function with different inputs
    # and outputs
    @pytest.mark.parametrize(
        "y_true, y_pred, expected",
        [
            (np.array([1, 2, 3]), np.array([1, 2, 3]), 0.0),  # exact match
            # one unit difference
            (np.array([1, 2, 3]), np.array([2, 3, 4]), 1.0),
            (
                    np.array([1, 2, 3]),
                    np.array([0, 0, 0]),
                    2,
            ),  # large difference
        ],
    )
    def test_mae_function(self, y_true, y_pred, expected, mae: MeanAbsoluteError):
        # Compare the expected and actual outputs with a precision of 7
        # decimal places
        np.testing.assert_almost_equal(
            mae.function(y_true, y_pred), expected, decimal=7
        )

    # Test the derivative method of the MAE loss function with different
    # inputs and outputs
    @pytest.mark.parametrize(
        "y_true, y_pred, expected",
        [
            (
                    np.array([1, 2, 3]),
                    np.array([1, 2, 3]),
                    np.array([0, 0, 0]),
            ),  # exact match
            (
                    np.array([1, 2, 3]),
                    np.array([2, 3, 4]),
                    np.array([0.33333333, 0.33333333, 0.33333333]),
            ),  # one unit difference
            (
                    np.array([1, 2, 3]),
                    np.array([0, 0, 0]),
                    np.array([-0.33333333, -0.33333333, -0.33333333]),
            ),  # large difference
        ],
    )
    def test_mae_derivative(self, y_true, y_pred, expected, mae: MeanAbsoluteError):
        # Compare the expected and actual outputs with a precision of 7
        # decimal places
        np.testing.assert_almost_equal(
            mae.gradient(y_true, y_pred), expected, decimal=7
        )



import pytest
import numpy as np

from inventronet.preprocessing import CountVectorizer


@pytest.fixture
def documents():
    return [
        "This is a simple document.",
        "This document is about programming.",
        "Another document is related to Python."
    ]


@pytest.fixture
def expected_vocabulary():
    return {
        'this': 0,
        'is': 1,
        'a': 2,
        'simple': 3,
        'document': 4,
        'about': 5,
        'programming': 6,
        'another': 7,
        'related': 8,
        'to': 9,
        'python': 10
    }


def test_count_vectorizer_fit_transform_vocabulary(documents, expected_vocabulary):
    vectorizer = CountVectorizer()
    vectorizer.fit_transform(documents)
    assert vectorizer.vocabulary_ == expected_vocabulary


def test_count_vectorizer_fit_transform_output(documents, expected_vocabulary):
    vectorizer = CountVectorizer()
    X = vectorizer.fit_transform(documents)

    expected_output = np.array([
        [1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0],
        [1, 1, 0, 0, 1, 1, 1, 0, 0, 0, 0],
        [0, 1, 0, 0, 1, 0, 0, 1, 1, 1, 1]
    ])

    assert np.array_equal(X, expected_output)


def test_count_vectorizer_transform_shape(documents, expected_vocabulary):
    vectorizer = CountVectorizer()
    vectorizer.fit(documents)

    new_documents = [
        "A new document about Python.",
        "This is a simple test."
    ]

    X_new = vectorizer.transform(new_documents)
    assert X_new.shape == (2, len(expected_vocabulary))


def test_count_vectorizer_lowercasing_shape(documents):
    vectorizer = CountVectorizer()
    X = vectorizer.fit_transform(documents)
    assert X.shape == (3, len(vectorizer.vocabulary_))


def test_count_vectorizer_no_lowercasing_shape(documents):
    vectorizer_no_lowercase = CountVectorizer(lowercase=False)
    X_no_lowercase = vectorizer_no_lowercase.fit_transform(documents)
    assert X_no_lowercase.shape == (3, len(vectorizer_no_lowercase.vocabulary_))


def test_count_vectorizer_no_lowercasing_vocabulary_size():
    documents = [
        "This is a Simple document.",
        "This DOCUMENT is about programming.",
        "Another document is related to Python."
    ]

    vectorizer = CountVectorizer()
    vectorizer.fit_transform(documents)

    vectorizer_no_lowercase = CountVectorizer(lowercase=False)
    vectorizer_no_lowercase.fit_transform(documents)

    assert len(vectorizer_no_lowercase.vocabulary_) > len(vectorizer.vocabulary_)




# Define a class for a sequential model
from typing import List, Tuple, Type

import numpy as np
from tqdm import tqdm

from .model import Model
from ..layers.layer import Layer
from ..losses.loss import Loss
from ..metrics.metric import Metric
from ..optimizers.optimizer import Optimizer


class Sequential(Model):
    # Define the initialization method
    def __init__(self) -> None:
        """Call the superclass constructor."""
        # Call the superclass constructor
        super().__init__()
        self.metrics: List[Type[Metric]] = None
        self.optimizer: Type[Optimizer] = None
        self.loss: Type[Loss] = None
        # Attributes for early stopping
        self.patience: int = None
        self.min_delta: float = None
        self.wait: int = 0
        self.best_loss: float = np.inf

        # Attribute for history
        self.history = {
            "loss": [],
        }

    # Add a method to set early stopping parameters
    def set_early_stopping(self, patience: int, min_delta: float) -> None:
        self.patience = patience
        self.min_delta = min_delta

    # Define a method for adding a layer to the model
    def add(self, layer: Type[Layer]) -> None:
        """Add a layer to the model.

        Args:
            layer (Layer): The layer to be added.

        Raises:
            AssertionError: If the layer input dimension is not None and
            does not match the previous layer output dimension.
        """
        # Check if the layer is compatible with the previous layer
        if self.layers:
            previous_layer = self.layers[-1]
            # If the layer input dimension is None, infer it from the
            # previous layer output dimension
            if layer.input_dim is None:
                layer.input_dim = previous_layer.output_dim
            # Otherwise, check if the layer input dimension matches the
            # previous layer output dimension
            else:
                assert (
                        layer.input_dim == previous_layer.output_dim
                ), "Layer input dimension does not match previous layer output dimension"
        # Append the layer to the list of layers
        self.layers.append(layer)

    def compile(
            self, loss: Type[Loss], optimizer: Type[Optimizer], metrics: List[Type[Metric]]
    ) -> None:
        """Compile the model with a loss function, an optimizer, and metrics.

        Args:
            loss (Loss): The loss function to be used.
            optimizer (Optimizer): The optimizer to be used.
            metrics (List[Metric]): The metrics to be used.
        """
        self.loss = loss
        self.optimizer = optimizer
        self.metrics = metrics
        # Initialize metric history
        for metric in metrics:
            self.history[f"{metric.__class__.__name__}"] = []

    def fit(
            self,
            x_train: np.ndarray,
            y_train: np.ndarray,
            epochs: int,
    ) -> None:
        """Fit the model on training data.

        Args:
            x_train (np.ndarray): The input data for training.
            y_train (np.ndarray): The output data for training.
            epochs (int): The number of epochs to train the model.
        """
        progress_bar = tqdm(range(epochs), desc="Training progress")
        for epoch in progress_bar:
            # Forward pass the input data through the network
            layer_output = x_train
            for layer in self.layers:
                layer_output = layer.forward(layer_output, training=True)

            # Calculate the loss and the metrics
            loss_value = self.loss.function(y_train, layer_output)
            metric_values = [
                metric.call(y_train, layer_output) for metric in self.metrics
            ]
            # Update the history
            self.history["loss"].append(loss_value)
            for metric, m_value in zip(self.metrics, metric_values):
                self.history[f"{metric.__class__.__name__}"].append(m_value)

            # Update the progress bar description with the loss and metrics
            metrics_str = ", ".join(
                [
                    f"{metric.__class__.__name__}: {m:.4f}"
                    for metric, m in zip(self.metrics, metric_values)
                ]
            )
            progress_bar.set_description(f"Epoch {epoch + 1}, Loss: {loss_value:.4f}, {metrics_str}")

            # Backward pass the error through the network
            layer_error = self.loss.gradient(y_train, layer_output)
            for layer_index, layer in enumerate(reversed(self.layers)):
                layer_input = (
                    layer.previous_layer_output
                    if layer.previous_layer_output is not None
                    else x_train
                )
                layer_error = layer.backward(layer_error, prev_output=layer_input)

                # Check shapes before updating the optimizer
                for key in layer.parameters.keys():
                    assert layer.parameters[key].shape == layer.gradients[
                        key].shape, f"Parameter shape {layer.parameters[key].shape} does not match gradient shape {layer.gradients[key].shape}"

                self.optimizer.update(len(self.layers) - 1 - layer_index, layer.parameters, layer.gradients)

            # Check for early stopping
            if self.patience is not None and self.min_delta is not None:
                if loss_value < self.best_loss - self.min_delta:
                    self.best_loss = loss_value
                    self.wait = 0
                else:
                    self.wait += 1

                if self.wait >= self.patience:
                    progress_bar.close()
                    print(f"Early stopping on epoch {epoch + 1}")
                    break

    def predict(self, x_test: np.ndarray) -> np.ndarray:
        # Forward pass the input data through the network
        layer_output = x_test
        for layer in self.layers:
            layer_output = layer.forward(layer_output)
        # Return the final output
        return layer_output

    # Define a method for evaluating the model on test data
    def evaluate(
            self, x_test: np.ndarray, y_test: np.ndarray
    ) -> Tuple[float, List[float]]:
        # Predict the output for the test data
        y_pred = self.predict(x_test)
        # Calculate the loss and the metrics
        loss_value = self.loss.function(y_test, y_pred)
        metric_values = [metric.call(y_test, y_pred) for metric in self.metrics]
        # Return the loss and the metrics
        return loss_value, metric_values



from typing import Optional
import numpy as np
from .activation import Activation


class ReLU(Activation):
    def function(self, x: np.ndarray) -> np.ndarray:
        """
        A method for the ReLU function.

        This method implements the ReLU function using the np.maximum
        function.

        Args:
            x: A numpy array of any shape, representing the input to the
            ReLU function.

        Returns:
            A numpy array of the same shape as x, representing the output of
            the ReLU function.
        """
        return np.maximum(x, 0)

    def derivative(
        self, x: np.ndarray, output: Optional[np.ndarray] = None
    ) -> np.ndarray:
        """
        A method for the derivative of the ReLU function.

        This method implements the derivative of the ReLU function using
        the np.where function.

        Args:
            x: A numpy array of any shape, representing the input to the
            ReLU function.

        Returns:
            A numpy array of the same shape as x, representing the output of
            the derivative of the ReLU function.
        """
        return np.where(x > 0, 1, 0)



import numpy as np
import pytest

from inventronet.losses import BinaryCrossEntropy


class TestBCE:
    # Create an instance of the BCE loss function
    @pytest.fixture
    def bce(self) -> BinaryCrossEntropy:
        yield BinaryCrossEntropy()

    # Test the function method of the BCE loss function with different
    # inputs and outputs
    @pytest.mark.parametrize(
        "y_true, y_pred, expected",
        [
            (
                    np.array([1, 0, 1]),
                    np.array([0.9, 0.1, 0.9]),
                    0.10536052,
            ),  # high confidence
            (
                    np.array([1, 0, 1]),
                    np.array([0.6, 0.4, 0.6]),
                    0.51082562,
            ),  # medium confidence
            (
                    np.array([1, 0, 1]),
                    np.array([0.5, 0.5, 0.5]),
                    0.69314718,
            ),  # low confidence
        ],
    )
    def test_bce_function(self, y_true, y_pred, expected, bce: BinaryCrossEntropy):
        # Compare the expected and actual outputs with a precision o
        # f 7 decimal places
        np.testing.assert_almost_equal(
            bce.function(y_true, y_pred), expected, decimal=7
        )

    # Test the derivative method of the BCE loss function with different
    # inputs and outputs
    @pytest.mark.parametrize(
        "y_true, y_pred, expected",
        [
            (
                    np.array([1, 0, 1]),
                    np.array([0.9, 0.1, 0.9]),
                    np.array([-0.37037037, 0.3703704, -0.37037037]),
            ),  # high confidence
            (
                    np.array([1, 0, 1]),
                    np.array([0.6, 0.4, 0.6]),
                    np.array([-0.55555556, 0.55555556, -0.55555556]),
            ),  # medium confidence
            (
                    np.array([1, 0, 1]),
                    np.array([0.5, 0.5, 0.5]),
                    np.array([-0.66666667, 0.66666667, -0.66666667]),
            ),  # low confidence
        ],
    )
    def test_bce_derivative(self, y_true, y_pred, expected, bce: BinaryCrossEntropy):
        # Compare the expected and actual outputs with a precision of 7
        # decimal places
        np.testing.assert_almost_equal(
            bce.gradient(y_true, y_pred), expected, decimal=7
        )



from abc import ABC, abstractmethod
import numpy as np


class Loss(ABC):
    @abstractmethod
    def function(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        """
        Computes the loss between the true labels and the predicted values.

        Parameters
        ----------
        y_true : np.ndarray
            The true labels of the data, encoded as one-hot vectors or as
                scalars.
            Shape: (n_samples, n_classes) or (n_samples,)
        y_pred : np.ndarray
            The predicted values of the data, output of the activation
                function.
            Shape: (n_samples, n_classes) or (n_samples,)

        Returns
        -------
        float
            The loss, averaged over all samples and classes or samples.
        """
        raise NotImplementedError("The function method must be implemented.")

    @abstractmethod
    def gradient(self, y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
        """
        Computes the derivative of the loss with respect to the predictions.

        Parameters
        ----------
        y_true : np.ndarray
            The true labels of the data, encoded as one-hot vectors or as
                scalars.
            Shape: (n_samples, n_classes) or (n_samples,)
        y_pred : np.ndarray
            The predicted values of the data, output of the activation
                function.
            Shape: (n_samples, n_classes) or (n_samples,)

        Returns
        -------
        np.ndarray
            The derivative of the loss, scaled by the number of samples.
            Shape: (n_samples, n_classes) or (n_samples,)
        """
        raise NotImplementedError("The gradient method must be implemented.")

    def __call__(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        return self.function(y_true, y_pred)



from typing import Optional
import numpy as np
from .activation import Activation


class Sigmoid(Activation):
    # Implement the abstract methods
    def function(self, x: np.ndarray) -> np.ndarray:
        """
        A method for the sigmoid function.

        This method implements the sigmoid function: 1 / (1 + exp(-x))

        Args:
            x: A numpy array of any shape, representing the input to
            the sigmoid function.

        Returns:
            A numpy array of the same shape as x, representing the output
            of the sigmoid function.
        """
        return 1 / (1 + np.exp(-x))

    def derivative(
        self, x: np.ndarray, output: Optional[np.ndarray] = None
    ) -> np.ndarray:
        """
        A method for the derivative of the sigmoid function.

        This method implements the derivative of the sigmoid function:
        sigmoid(x) * (1 - sigmoid(x))

        Args:
            x: A numpy array of any shape, representing the input to the
            sigmoid function.

        Returns:
            A numpy array of the same shape as x, representing the output of
            the derivative of the sigmoid function.
        """
        if output is None:
            output = self.function(x)
        return output * (1 - output)
